#!/usr/bin/env python
# -*- coding: utf-8 -*-
# filename: __rpc.py

# Tencent is pleased to support the open source community by making Tars available.
#
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
#


'''
@version: 0.01
@brief: rpc调用逻辑实现
'''

import argparse
import time

from .__TimeoutQueue import QueueTimeout
from .__TimeoutQueue import TimeoutQueue
from .__adapterproxy import AdapterProxyManager
from .__async import AsyncProcThread
from .__logger import initLog
from .__logger import tarsLogger
from .__servantproxy import ServantProxy
from .__trans import EndPointInfo
from .__trans import FDReactor
from .exception import (TarsException)


class Communicator:
    '''
    @brief: 通讯器，创建和维护ServantProxy、ObjectProxy、FDReactor线程和超时线程
    '''
    default_config = {'tars':
                      {'application':
                       {'client':
                        {'async-invoke-timeout': 20000,
                         'asyncthread': 0,
                         'locator': '',
                            'loglevel': 'error',
                            'logpath': 'tars.log',
                            'logsize': 15728640,
                            'lognum': 0,
                            'refresh-endpoint-interval': 60000,
                            'sync-invoke-timeout': 5000}}}}

    def __init__(self, config={}):
        tarsLogger.debug('Communicator:__init__')
        self.__terminate = False
        self.__initialize = False
        self.__objects = {}
        self.__servants = {}
        self.__reactor = None
        self.__qTimeout = None
        self.__asyncProc = None
        self.__config = Communicator.default_config.copy()
        self.__config.update(config)
        self.initialize()

    def __del__(self):
        tarsLogger.debug('Communicator:__del__')

    def initialize(self):
        '''
        @brief: 使用通讯器前必须先调用此函数
        '''
        tarsLogger.debug('Communicator:initialize')
        if self.__initialize:
            return
        logpath = self.getProperty('logpath')
        logsize = self.getProperty('logsize', int)
        lognum = self.getProperty('lognum', int)
        loglevel = self.getProperty('loglevel')
        initLog(logpath, logsize, lognum, loglevel)

        self.__reactor = FDReactor()
        self.__reactor.initialize()
        self.__reactor.start()

        self.__qTimeout = QueueTimeout()
        self.__qTimeout.setHandler(self.handleTimeout)
        self.__qTimeout.start()

        async_num = self.getProperty('asyncthread', int)
        self.__asyncProc = AsyncProcThread()
        self.__asyncProc.initialize(async_num)
        self.__asyncProc.start()

        self.__initialize = True

    def terminate(self):
        '''
        @brief: 不再使用通讯器需调用此函数释放资源
        '''
        tarsLogger.debug('Communicator:terminate')

        if not self.__initialize:
            return

        self.__reactor.terminate()
        self.__qTimeout.terminate()
        self.__asyncProc.terminate()

        for objName in self.__servants:
            self.__servants[objName]._terminate()

        for objName in self.__objects:
            self.__objects[objName].terminate()

        self.__objects = {}
        self.__servants = {}
        self.__reactor = None
        self.__initialize = False

    def parseConnAddr(self, connAddr):
        '''
        @brief: 解析connAddr字符串
        @param connAddr: 连接地址
        @type connAddr: str
        @return: 解析结果
        @rtype: dict, key是str，val里name是str，
                timeout是float，endpoint是EndPointInfo的list
        '''
        tarsLogger.debug('Communicator:parseConnAddr')
        connAddr = connAddr.strip()
        connInfo = {
            'name': '',
            'timeout': -1,
            'endpoint': []
        }
        if '@' not in connAddr:
            connInfo['name'] = connAddr
            return connInfo

        try:
            tks = connAddr.split('@')
            connInfo['name'] = tks[0]
            tks = tks[1].lower().split(':')
            parser = argparse.ArgumentParser(add_help=False)
            parser.add_argument('-h')
            parser.add_argument('-p')
            parser.add_argument('-t')
            for tk in tks:
                argv = tk.split()
                if argv[0] != 'tcp':
                    raise TarsException(
                        'unsupport transmission protocal : %s' % connInfo['name'])
                mes = parser.parse_args(argv[1:])
                try:
                    ip = mes.h if mes.h is not None else ''
                    port = int(mes.p) if mes.p is not None else '-1'
                    timeout = int(mes.t) if mes.t is not None else '-1'
                    connInfo['endpoint'].append(
                        EndPointInfo(ip, port, timeout))
                except Exception:
                    raise TarsException('Unrecognized option : %s' % mes)
        except TarsException:
            raise

        except Exception as exp:
            raise TarsException(exp)

        return connInfo

    def getReactor(self):
        '''
        @brief: 获取reactor
        '''
        return self.__reactor

    def getAsyncProc(self):
        '''
        @brief: 获取asyncProc
        '''
        return self.__asyncProc

    def getProperty(self, name, dt_type=str):
        '''
        @brief: 获取配置
        @param name: 配置名称
        @type name: str
        @param dt_type: 数据类型
        @type name: type
        @return: 配置内容
        @rtype: str
        '''
        try:
            ret = self.__config['tars']['application']['client'][name]
            ret = dt_type(ret)
        except:
            ret = Communicator.default_config['tars']['application']['client'][name]

        return ret

    def setProperty(self, name, value):
        '''
        @brief: 修改配置
        @param name: 配置名称
        @type propertys: str
        @param value: 配置内容
        @type propertys: str
        @return: 设置是否成功
        @rtype: bool
        '''
        try:
            self.__config['tars']['application']['client'][name] = value
            return True
        except:
            return False

    def setPropertys(self, propertys):
        '''
        @brief: 修改配置
        @param propertys: 配置集合
        @type propertys: map, key type: str, value type: str
        @return: 无
        @rtype: None
        '''
        pass

    def updateConfig(self):
        '''
        @brief: 重新设置配置
        '''

    def stringToProxy(self, servantProxy, connAddr):
        '''
        @brief: 初始化ServantProxy
        @param connAddr: 服务器地址信息
        @type connAddr: str
        @param servant: servant proxy
        @type servant: ServantProxy子类
        @return: 无
        @rtype: None
        @note: 如果connAddr的ServantObj连接过，返回连接过的ServantProxy
               如果没有连接过，用参数servant初始化，返回servant
        '''
        tarsLogger.debug('Communicator:stringToProxy')

        connInfo = self.parseConnAddr(connAddr)
        objName = connInfo['name']
        if objName in self.__servants:
            return self.__servants[objName]

        objectPrx = ObjectProxy()
        objectPrx.initialize(self, connInfo)

        servantPrx = servantProxy()
        servantPrx._initialize(self.__reactor, objectPrx)
        self.__objects[objName] = objectPrx
        self.__servants[objName] = servantPrx
        return servantPrx

    def handleTimeout(self):
        '''
        @brief: 处理超时事件
        @return: 无
        @rtype: None
        '''
        # tarsLogger.debug('Communicator:handleTimeout')
        for obj in self.__objects.values():
            obj.handleQueueTimeout()


class ObjectProxy:
    '''
    @brief: 一个object name在一个Communicator里有一个objectproxy
            管理收发的消息队列
    '''
    DEFAULT_TIMEOUT = 3.0

    def __init__(self):
        tarsLogger.debug('ObjectProxy:__init__')
        self.__name = ''
        self.__timeout = ObjectProxy.DEFAULT_TIMEOUT
        self.__comm = None
        self.__epi = None
        self.__adpmanager = None
        self.__timeoutQueue = None
        # self.__adapter = None
        self.__initialize = False

    def __del__(self):
        tarsLogger.debug('ObjectProxy:__del__')

    def initialize(self, comm, connInfo):
        '''
        @brief: 初始化，使用ObjectProxy前必须调用
        @param comm: 通讯器
        @type comm: Communicator
        @param connInfo: 连接信息
        @type comm: dict
        @return: None
        @rtype: None
        '''
        if self.__initialize:
            return
        tarsLogger.debug('ObjectProxy:initialize')
        self.__comm = comm
        # async-invoke-timeout来设置队列时间
        async_timeout = self.__comm.getProperty(
            'async-invoke-timeout', float) / 1000
        self.__timeoutQueue = TimeoutQueue(async_timeout)

        self.__name = connInfo['name']

        self.__timeout = self.__comm.getProperty(
            'sync-invoke-timeout', float) / 1000

        # 通过Communicator的配置设置超时
        # 不再通过连接信息的-t来设置
        # if connInfo['timeout'] != -1:
        # self.__timeout = connInfo['timeout']
        eplist = connInfo['endpoint']

        self.__adpmanager = AdapterProxyManager()
        self.__adpmanager.initialize(comm, self, eplist)

        self.__initialize = True

    def terminate(self):
        '''
        @brief: 回收资源，不再使用ObjectProxy时调用
        @return: None
        @rtype: None
        '''
        tarsLogger.debug('ObjectProxy:terminate')
        self.__timeoutQueue = None
        self.__adpmanager.terminate()
        self.__initialize = False

    def name(self):
        '''
        @brief: 获取object name
        @return: object name
        @rtype: str
        '''
        return self.__name

    # def setTimeout(self, timeout):
        # '''
        # @brief: 设置超时
        # @param timeout: 超时时间，单位为s
        # @type timeout: float
        # @return: None
        # @rtype: None
        # '''
        # self.__timeout = timeout
        # self.__timeoutQueue.setTimeout(timeout)

    def timeout(self):
        '''
        @brief: 获取超时时间
        @return: 超时时间，单位为s
        @rtype: float
        '''
        return self.__timeout

    def getTimeoutQueue(self):
        '''
        @brief: 获取超时队列
        @return: 超时队列
        @rtype: TimeoutQueue
        '''
        return self.__timeoutQueue

    def handleQueueTimeout(self):
        '''
        @brief: 超时事件发生时处理超时事务
        @return: None
        @rtype: None
        '''
        # tarsLogger.debug('ObjectProxy:handleQueueTimeout')
        self.__timeoutQueue.timeout()

    def invoke(self, reqmsg):
        '''
        @brief: 远程过程调用
        @param reqmsg: 请求响应报文
        @type reqmsg: ReqMessage
        @return: 错误码
        @rtype:
        '''
        tarsLogger.debug('ObjectProxy:invoke, objname: %s, func: %s',
                         self.__name, reqmsg.request.sFuncName)
        # 负载均衡
        # adapter = self.__adpmanager.getNextValidProxy()
        adapter = self.__adpmanager.selectAdapterProxy(reqmsg)
        if not adapter:
            tarsLogger.error("invoke %s, select adapter proxy return None",
                             self.__name)
            return -2

        adapter.checkActive(True)
        reqmsg.adapter = adapter
        return adapter.invoke(reqmsg)

    # 弹出请求报文
    def popRequest(self):
        '''
        @brief: 返回消息队列里的请求响应报文，FIFO
                不删除TimeoutQueue里的数据，响应时要用
        @return: 请求响应报文
        @rtype: ReqMessage
        '''
        return self.__timeoutQueue.pop(erase=False)


if __name__ == '__main__':
    connAddr = "apptest.lightServer.lightServantObj@tcp -h 10.130.64.220 -p 10001 -t 10000"
    connAddr = 'MTT.BookMarksUnifyServer.BookMarksUnifyObj@tcp -h 172.17.149.77 -t 60000 -p 10023'
    comm = Communicator()
    comm.initialize()
    servant = ServantProxy()
    servant = comm.stringToProxy(connAddr, servant)
    print(servant.tars_timeout())
    try:
        rsp = servant.tars_invoke(
            ServantProxy.TARSNORMAL, "test", '', ServantProxy.mapcls_context(), None)
        print('Servant invoke success, request id: %d, iRet: %d' % (
            rsp.iRequestId, rsp.iRet))
    except Exception as msg:
        print(msg)
    finally:
        servant.tars_terminate()
    time.sleep(2)
    print('app closing ...')
    comm.terminate()
    time.sleep(2)
    print('cpp closed')



from biliup.plugins.huya_wup import Wup, DEFAULT_TICKET_NUMBER
from biliup.plugins.huya_wup.packet import (
    HuyaGetCdnTokenReq,
    HuyaGetCdnTokenRsp
)
from biliup.plugins.huya_wup.packet.__util import STANDARD_CHARSET

import httpx
client = httpx.Client()

if __name__ == "__main__":
    # import base64
    cdn: str = "TX"
    stream_name: str = "1199627305549-1199627305549-5718448156589424640-2399254734554-10057-A-0-1-imgplus.flv"
    presenter_uid: int = 1199627305549
    wup_url: str = "https://wup.huya.com"
    huya_headers: dict = {
        "user-agent": f"HYSDK(Windows, 30000002)_APP(pc_exe&6090007&official)_SDK(trans&2.24.0.5157)",
        "referer": "https://www.huya.com/",
        "origin": "https://www.huya.com",
    }
    # TupVersion3 = 3
    wup_req = Wup()
    # wup_req.version = TupVersion3
    wup_req.requestid = abs(DEFAULT_TICKET_NUMBER)
    wup_req.servant = "liveui"
    wup_req.func = "getCdnTokenInfo"
    token_info_req = HuyaGetCdnTokenReq()
    token_info_req.cdnType = cdn
    token_info_req.streamName = stream_name
    token_info_req.presenterUid = presenter_uid
    wup_req.put(
        vtype=HuyaGetCdnTokenRsp,
        name="tReq",
        value=token_info_req
    )
    data = wup_req.encode_v3()
    rsp = client.post(wup_url, data=data, headers=huya_headers)
    rsp_bytes = rsp.content
    wup_rsp = Wup()
    wup_rsp.decode_v3(rsp_bytes)
    token_info_rsp = wup_rsp.get(
        vtype=HuyaGetCdnTokenRsp,
        # name=bytes("tRsp", encoding=STANDARD_CHARSET)
        name="tRsp"
    )
    print(token_info_rsp.as_dict())



#! /usr/bin/env python
# -*- coding: utf-8 -*-


# Tencent is pleased to support the open source community by making Tars available.
#
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
#

__version__ = "0.0.1"

from biliup.Danmaku.tars.__tars import TarsInputStream
from biliup.Danmaku.tars.__tars import TarsOutputStream
from biliup.Danmaku.tars.__tup import TarsUniPacket
from biliup.Danmaku.tars.__util import util


class tarscore:
    class TarsInputStream(TarsInputStream):
        pass

    class TarsOutputStream(TarsOutputStream):
        pass

    class TarsUniPacket(TarsUniPacket):
        pass

    class boolean(util.boolean):
        pass

    class int8(util.int8):
        pass

    class uint8(util.uint8):
        pass

    class int16(util.int16):
        pass

    class uint16(util.uint16):
        pass

    class int32(util.int32):
        pass

    class uint32(util.uint32):
        pass

    class int64(util.int64):
        pass

    class float(util.float):
        pass

    class double(util.double):
        pass

    class bytes(util.bytes):
        pass

    class string(util.string):
        pass

    class struct(util.struct):
        pass

    @staticmethod
    def mapclass(ktype, vtype): return util.mapclass(ktype, vtype)

    @staticmethod
    def vctclass(vtype): return util.vectorclass(vtype)

    @staticmethod
    def printHex(buff): util.printHex(buff)


# 被用户引用



import json
import re

import biliup.common.util
from . import logger
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase


@Plugin.download(regexp=r'(?:https?://)?(?:(?:www)\.)?inke\.cn')
class Inke(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)

    async def acheck_stream(self, is_check=False):
        logger.debug(self.fname)
        rid = re.search(r'uid=([a-zA-Z0-9]+)', self.url).group(1)
        r1 = await biliup.common.util.client.get(
            f"https://webapi.busi.inke.cn/web/live_share_pc?uid={rid}",
            timeout = 5,
            headers = self.fake_headers
        )
        # r1.close()
        jsons = json.loads(r1.text)
        if jsons:
            if jsons.get('error_code') != 0:
                logger.error("直播间地址错误")
                return False
            if jsons['data']['status']:
                self.room_title = jsons['data']['live_name']
                self.raw_stream_url = jsons['data']['live_addr'][0]['stream_addr']
                return True

        logger.debug("主播未开播")
        return False



import time
import json
import re
import asyncio

from biliup.common.util import client
from biliup.config import config
from . import match1, logger, wbi
from biliup.Danmaku import DanmakuClient
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase


OFFICIAL_API = "https://api.live.bilibili.com"
STREAM_NAME_REGEXP = r"/live-bvc/\d+/(live_[^/\.]+)"
WBI_WEB_LOCATION = "444.8"

@Plugin.download(regexp=r'https?://(b23\.tv|live\.bilibili\.com)')
class Bililive(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)
        self.live_start_time = 0
        self.bilibili_danmaku = config.get('bilibili_danmaku', False)
        self.bilibili_danmaku_detail = config.get('bilibili_danmaku_detail', False)
        self.bilibili_danmaku_raw = config.get('bilibili_danmaku_raw', False)
        self.__real_room_id = None
        self.__login_mid = 0
        self.__anchor_mid = 0
        self.bili_cookie = config.get('user', {}).get('bili_cookie')
        self.bili_cookie_file = config.get('user', {}).get('bili_cookie_file')
        self.bili_qn = int(config.get('bili_qn', 25000))
        self.bili_protocol = config.get('bili_protocol', 'stream')
        self.bili_cdn = config.get('bili_cdn', [])
        self.bili_hls_timeout = config.get('bili_hls_transcode_timeout', 60)
        self.bili_api_list = [
            normalize_url(config.get('bili_liveapi', OFFICIAL_API)).rstrip('/'),
            normalize_url(config.get('bili_fallback_api', OFFICIAL_API)).rstrip('/'),
        ]
        self.bili_anonymous_origin = config.get('bili_anonymous_origin', False)
        self.bili_cdn_fallback = config.get('bili_cdn_fallback', False)

    async def acheck_stream(self, is_check=False):

        if "b23.tv" in self.url:
            try:
                resp = await client.get(self.url, follow_redirects=False)
                if resp.status_code not in {301, 302}:
                    raise Exception("不支持的链接")
                url = str(resp.next_request.url)
                if "live.bilibili" not in url:
                    raise Exception("不支持的链接")
                self.url = url
            except Exception as e:
                logger.error(f"{self.plugin_msg}: {e}")
                return False

        room_id: str = match1(self.url, r'bilibili.com/(\d+)')
        if self.bili_cookie:
            self.fake_headers['cookie'] = self.bili_cookie
        if self.bili_cookie_file:
            try:
                with open(self.bili_cookie_file, encoding='utf-8') as stream:
                    cookies = json.load(stream)["cookie_info"]["cookies"]
                    cookies_str = ''
                    for i in cookies:
                        cookies_str += f"{i['name']}={i['value']};"
                    self.fake_headers['cookie'] = cookies_str
            except Exception:
                logger.exception("load_cookies error")
        self.fake_headers['referer'] = self.url

        if int(time.time()) - wbi.last_update >= wbi.UPDATE_INTERVAL:
            await self.update_wbi()

        # 获取直播状态与房间标题
        try:
            params = {
                "room_id": room_id,
                "web_location": WBI_WEB_LOCATION,
            }
            wbi.sign(params)
            room_info = await client.get(
                f"{OFFICIAL_API}/xlive/web-room/v1/index/getInfoByRoom",
                params=params,
                headers=self.fake_headers)
            room_info.raise_for_status()
            room_info = room_info.json()
        except Exception as e:
            logger.error(f"{self.plugin_msg}: {e}", exc_info=True)
            return False
        if room_info['code'] != 0:
            logger.error(f"{self.plugin_msg}: {room_info}")
            return False
        else:
            room_info = room_info['data']
        if room_info['room_info']['live_status'] != 1:
            logger.debug(f"{self.plugin_msg}: 未开播")
            self.raw_stream_url = None
            return False

        self.live_cover_url = room_info['room_info']['cover']
        self.room_title = room_info['room_info']['title']
        self.__real_room_id = room_info['room_info']['room_id']
        self.__anchor_mid = room_info['room_info']['uid']
        live_start_time = room_info['room_info']['live_start_time']
        special_type = room_info['room_info']['special_type'] # 0: 公开直播, 1: 付费直播, 199: 纯净页面
        if live_start_time > self.live_start_time:
            self.live_start_time = live_start_time
            is_new_live = True
        else:
            is_new_live = False

        if is_check:
            return True
        else:
            self.__login_mid = await self.check_login_status()

        # 复用原画 m3u8 流
        if  self.raw_stream_url is not None \
            and ".m3u8" in self.raw_stream_url \
            and self.bili_qn >= 10000 \
            and not is_new_live:
            url = await self.acheck_url_healthy(self.raw_stream_url)
            if url is not None:
                logger.debug(f"{self.plugin_msg}: 复用 {url}")
                return True
            else:
                self.raw_stream_url = None


        stream_urls = await self.aget_stream(self.bili_qn, self.bili_protocol, special_type)
        if not stream_urls:
            if self.bili_protocol == 'hls_fmp4':
                if int(time.time()) - live_start_time <= self.bili_hls_timeout:
                    logger.warning(f"{self.plugin_msg}: 暂未提供 hls_fmp4 流，等待下一次检测")
                    return False
                else:
                    # 回退首个可用格式
                    stream_urls = await self.aget_stream(self.bili_qn, 'stream', special_type)
            else:
                logger.error(f"{self.plugin_msg}: 获取{self.bili_protocol}流失败")
                return False

        target_quality_stream = stream_urls.get(
            self.bili_qn, next(iter(stream_urls.values()))
        )
        stream_url = {}
        if self.bili_cdn is not None:
            for cdn in self.bili_cdn:
                stream_info = target_quality_stream.get(cdn)
                if stream_info is not None:
                    current_cdn = cdn
                    stream_url = stream_info['url']
                    break
        if not stream_url:
            current_cdn, stream_info = next(iter(target_quality_stream.items()))
            stream_url = stream_info['url']
            logger.debug(f"{self.plugin_msg}: 使用 {current_cdn} 流")

        self.raw_stream_url = f"{stream_url['host']}{stream_url['base_url']}{stream_url['extra']}"

        # 回退
        if self.bili_cdn_fallback:
            __url = await self.acheck_url_healthy(self.raw_stream_url)
            if __url is None:
                for cdn, stream_info in target_quality_stream.items():
                    stream_url = stream_info['url']
                    __fallback_url = f"{stream_url['host']}{stream_url['base_url']}{stream_url['extra']}"
                    try:
                        __url = await self.acheck_url_healthy(__fallback_url)
                        if __url is not None:
                            self.raw_stream_url = __url
                            logger.info(f"{self.plugin_msg}: cdn_fallback 回退到 {cdn} - {__fallback_url}")
                            break
                    except Exception as e:
                        logger.error(f"{self.plugin_msg}: cdn_fallback {e} - {__fallback_url}")
                        continue
                else:
                    logger.error(f"{self.plugin_msg}: 所有 cdn 均不可用")
                    self.raw_stream_url = None
                    return False
            else:
                self.raw_stream_url = __url

        return True

    def danmaku_init(self):
        if self.bilibili_danmaku:
            self.danmaku = DanmakuClient(
                self.url, self.gen_download_filename(), {
                    'room_id': self.__real_room_id,
                    'cookie': self.fake_headers.get('cookie', ''),
                    'detail': self.bilibili_danmaku_detail,
                    'raw': self.bilibili_danmaku_raw,
                    'uid': self.__login_mid
                }
            )


    async def get_play_info(self, api: str, qn: int = 10000) -> dict:
        full_url = f"{api}/xlive/web-room/v2/index/getRoomPlayInfo"
        try:
            params = {
                'room_id': str(self.__real_room_id),
                # 'no_playurl': '0',
                # 'mask': '1',
                'qn': str(qn),
                'platform': 'html5',  # 平台名称，web, html5, android, ios
                'protocol': '0,1',  # 流协议，0: http_stream(flv), 1: http_hls
                'format': '0,1,2',  # 编码格式，0: flv, 1: ts, 2: fmp4
                'codec': '0',  # 编码器，0: avc, 1: hevc, 2: av1
                # 'ptype': '8', # P2P配置，-1: disable, 8: WebRTC, 8192: MisakaTunnel
                'dolby': '5', # 杜比格式，5: 杜比音频
                # 'panorama': '1', # 全景(不支持 html5)
                # 'hdr_type': '0,1', # HDR类型(不支持 html5)，0: SDR, 1: PQ
                # 'req_reason': '0', # 请求原因，0: Normal, 1: PlayError
                # 'http': '1', # 优先 http 协议
                'web_location': WBI_WEB_LOCATION,
            }
            wbi.sign(params)
            api_res = await client.get(
                full_url, params=params, headers=self.fake_headers
            )
            api_res = json.loads(api_res.text)
            if api_res['code'] != 0:
                logger.error(f"{self.plugin_msg}: {api} 返回内容错误: {api_res}")
                return {}
            return api_res['data']
        except json.JSONDecodeError:
            logger.error(f"{self.plugin_msg}: {api} 返回内容错误: {api_res.text}")
        except Exception as e:
            logger.error(f"{self.plugin_msg}: {api} 获取 play_info 失败 -> {e}", exc_info=True)
        return {}

    async def get_master_m3u8(self, api: str) -> dict:
        full_url = f"{api}/xlive/play-gateway/master/url"
        params = {
            "cid": self.__real_room_id,
            "mid": self.__login_mid or self.__anchor_mid,
            "pt": "web", # platform
            "p2p_type": "-1",
            "net": 0,
            "free_type": 0,
            "build": 0,
            "feature": 2,
            "qn": self.bili_qn,
            "drm_type": 0,
            "codec": "0,1",
        }
        try:
            m3u8_res = await client.get(
                full_url, params=params, headers=self.fake_headers
            )
            if m3u8_res.status_code == 200 and m3u8_res.text.startswith("#EXTM3U"):
                return self.parse_master_m3u8(m3u8_res.text)
        except Exception as e:
            logger.error(f"{self.plugin_msg}: {api} 获取 m3u8 失败 -> {e}", exc_info=True)
        return {}

    async def aget_stream(self, qn: int = 10000, protocol: str = 'stream', special_type: int = 0) -> dict:
        """
        :param qn: 目标画质
        :param protocol: 流协议
        :param special_type: 特殊直播类型
        :return: 流信息
        """
        stream_urls = {}
        for api in self.bili_api_list:
            play_info = await self.get_play_info(api, qn)
            if not play_info or check_areablock(play_info):
                # logger.error(f"{self.plugin_msg}: {api} 返回内容错误: {play_info}")
                continue
            streams = play_info['playurl_info']['playurl']['stream']
            if protocol == 'hls_fmp4':
                if self.bili_anonymous_origin:
                    if special_type in play_info['all_special_types'] and not self.__login_mid:
                        logger.warn(f"{self.plugin_msg}: 特殊直播{special_type}")
                    else:
                        stream_urls = await self.get_master_m3u8(api)
                        if stream_urls:
                            break
                # 处理 API 信息
                stream = streams[1] if len(streams) > 1 else streams[0]
                for format in stream['format']:
                    if format['format_name'] == 'fmp4':
                        stream_urls = self.parse_stream_url(format['codec'][0])
                        # fmp4 可能没有原画
                        if qn in {10000, 25000} and qn not in stream_urls.keys():
                            stream_urls = {}
            else:
                stream_urls = self.parse_stream_url(streams[0]['format'][0]['codec'][0])
            if stream_urls:
                break
        # 空字典照常返回，重试交给上层方法处理
        return stream_urls

    async def get_user_status(self) -> dict:
        try:
            nav_res = await client.get(
                'https://api.bilibili.com/x/web-interface/nav',
                headers=self.fake_headers
            )
            nav_res.raise_for_status()
            nav_res = json.loads(nav_res.text)
            if (
                nav_res['code'] == 0 or
                (nav_res['code'] == -101 and nav_res['message'] == '账号未登录')
            ):
                return nav_res['data']
            logger.error(f"{self.plugin_msg}: 获取 nav 失败-{nav_res}")
        except:
            logger.error(f"{self.plugin_msg}: 获取 nav 失败", exc_info=True)
        return {}

    async def update_wbi(self):
        def _extract_key(url):
            if not url:
                return None
            slash = url.rfind('/')
            dot = url.find('.', slash)
            if slash == -1 or dot == -1:
                return None
            return url[slash + 1:dot]
        data = await self.get_user_status()
        wbi_key = data.get('wbi_img')
        if wbi_key:
            img_key = _extract_key(wbi_key.get('img_url'))
            sub_key = _extract_key(wbi_key.get('sub_url'))
            if img_key and sub_key:
                wbi.update_key(img_key, sub_key)
            else:
                logger.warning(f"img_key-{img_key}, sub_key-{sub_key}")
        else:
            logger.warning(f"Can not get wbi key by {data}")

    async def check_login_status(self) -> int:
        """
        检查B站登录状态
        :return: 当前登录用户 mid
        """
        try:
            data = await self.get_user_status()
            if data.get('isLogin'):
                logger.info(f"用户名：{data['uname']}, mid: {data['mid']}")
                return data['mid']
            else:
                logger.warning(f"{self.plugin_msg}: 未登录，或将只能录制到最低画质。")
        except Exception as e:
            logger.error(f"{self.plugin_msg}: 登录态校验失败 {e}")
        return 0

    def parse_stream_url(self, *args) -> dict:
        suffix_regexp = r'suffix=([^&]+)'
        if isinstance(args[0], str):
            url = args[0]
            host = "https://" + match1(url, r'https?://([^/]+)')
            stream_url = {
                'host': host,
                'base_url': url.split("?")[0].split(host)[1] + "?",
                'extra': url.split("?")[1]
            }
            return {
                'url': stream_url,
                'stream_name': match1(url, STREAM_NAME_REGEXP),
                'suffix': match1(url, suffix_regexp)
            }
        elif isinstance(args[0], dict):
            streams = {}
            current_qn = args[0]['current_qn']
            streams.setdefault(current_qn, {})
            base_url = args[0]['base_url']
            for info in args[0]['url_info']:
                cdn_name = match1(info['extra'], r'cdn=([^&]+)')
                stream_url = {
                    'host': info['host'],
                    'base_url': base_url,
                    'extra': info['extra']
                }
                streams[current_qn].setdefault(cdn_name, {
                    'url': stream_url,
                    'stream_name': match1(base_url, STREAM_NAME_REGEXP),
                    'suffix': match1(info['extra'], suffix_regexp)
                })
            return streams


    def parse_master_m3u8(self, m3u8_content: str) -> dict:
        """
        Returns:
            {
                "qn值": {
                    "cdn名称": {
                        "url": parsed_stream_url,
                        "stream_name": "流名称",
                        "suffix": "二压后缀"
                    }
                }
            }
        """
        lines = m3u8_content.strip().splitlines()
        current_qn = None
        result = {}

        if not lines[0].startswith('#EXTM3U'):
            raise ValueError('Invalid m3u8 file')

        for line in lines:
            if line.startswith('#EXT-X-STREAM-INF:'):
                codec = match1(line, r'CODECS="([^"]+)"')
                current_qn = int(match1(line, r'BILI-QN=(\d+)'))

                if codec and current_qn:
                    if 'avc' in codec.lower():
                        result.setdefault(current_qn, {})
                    else:
                        current_qn = None

            elif line.startswith('http') and current_qn is not None:
                cdn_name = match1(line, r'cdn=([^&]+)')
                if cdn_name:
                    result[current_qn].setdefault(cdn_name, self.parse_stream_url(line))

        return dict(sorted(result.items(), key=lambda x: int(x[0]), reverse=True))

# Copy from room-player.js
def check_areablock(data):
    '''
    :return: True if area block
    '''
    if not data['playurl_info']['playurl']:
        logger.error('Sorry, bilibili is currently not available in your country according to copyright restrictions.')
        logger.error('非常抱歉，根据版权方要求，您所在的地区无法观看本直播')
        return True
    return False

def normalize_url(url: str) -> str:
    return url if url.startswith(('http://', 'https://')) else 'http://' + url


import logging
import random
import re

logger = logging.getLogger('biliup')


class Twitch:
    heartbeat = "PING"
    heartbeatInterval = 40

    @staticmethod
    async def get_ws_info(url, context):
        reg_datas = []
        room_id = re.search(r"/([^/?]+)[^/]*$", url).group(1)

        reg_datas.append("CAP REQ :twitch.tv/tags twitch.tv/commands twitch.tv/membership")
        reg_datas.append("PASS SCHMOOPIIE")
        nick = f"justinfan{int(8e4 * random.random() + 1e3)}"
        reg_datas.append(f"NICK {nick}")
        reg_datas.append(f"USER {nick} 8 * :{nick}")
        reg_datas.append(f"JOIN #{room_id}")

        return "wss://irc-ws.chat.twitch.tv", reg_datas

    @staticmethod
    def decode_msg(data):
        msgs = []
        if data is not None:
            for d in data.splitlines():
                msgt = {}
                try:
                    msgt["content"] = re.search(r"PRIVMSG [^:]+:(.+)", d).group(1)
                    msgt["name"] = re.search(r"display-name=([^;]+);", d).group(1)
                    # if msgt["content"][0] == '@': continue # 丢掉表情符号
                    c = re.search(r"color=#([a-zA-Z0-9]{6});", d).group(1)
                    msgt["color"] = int(c, 16)
                    msgt["msg_type"] = "danmaku"
                    # print(msgt)
                    msgs.append(msgt)
                except Exception:
                    pass
        return msgs



from biliup.common.tars import tarscore

class HuyaWSUserInfo(tarscore.struct):
    __tars_class__ = "Huya.WSUserInfo"

    def __init__(self):
        self.lUid: tarscore.int64 = 0
        self.bAnonymous: tarscore.boolean = True
        self.sGuid: tarscore.string = ""
        self.sToken: tarscore.string = ""
        self.lTid: tarscore.int64 = 0
        self.lSid: tarscore.int64 = 0
        self.lGroupId: tarscore.int64 = 0
        self.lGroupType: tarscore.int64 = 0

    @staticmethod
    def writeTo(oos: tarscore.TarsOutputStream, value):
        oos.write(tarscore.int64, 0, value.lUid)
        oos.write(tarscore.boolean, 1, value.bAnonymous)
        oos.write(tarscore.string, 2, value.sGuid)
        oos.write(tarscore.string, 3, value.sToken)
        oos.write(tarscore.int64, 4, value.lTid)
        oos.write(tarscore.int64, 5, value.lSid)
        oos.write(tarscore.int64, 6, value.lGroupId)
        oos.write(tarscore.int64, 7, value.lGroupType)

    @staticmethod
    def readFrom(ios: tarscore.TarsInputStream):
        value = HuyaWSUserInfo()
        value.lUid = ios.read(tarscore.int64, 0, False)
        value.bAnonymous = ios.read(tarscore.boolean, 1, False)
        value.sGuid = ios.read(tarscore.string, 2, False)
        value.sToken = ios.read(tarscore.string, 3, False)
        value.lTid = ios.read(tarscore.int64, 4, False)
        value.lSid = ios.read(tarscore.int64, 5, False)
        value.lGroupId = ios.read(tarscore.int64, 6, False)
        value.lGroupType = ios.read(tarscore.int64, 7, False)



"""
stream_gears 上传功能测试

这个测试文件展示了如何使用 stream_gears.upload() 函数的各种功能：
1. 基础上传 - 使用关键字参数让代码更清晰
2. 多文件上传 - 上传多P视频
3. 定时发布 - 设置未来时间发布
4. 高级功能 - 使用富文本简介和额外字段
5. 最小参数 - 只使用必需参数，其他使用默认值

注意：运行前请确保：
- 存在有效的 cookies.json 文件
- 视频文件路径正确
- 网络连接正常
"""

import stream_gears
from typing import List, Dict, Any

video_path = ["E:\\Projects\\biliup\\target\\x86_64-pc-windows-msvc\\debug\\test.flv"]
cookie_file = "E:\\Projects\\biliup\\target\\x86_64-pc-windows-msvc\\debug\\cookies.json"
# submit_api = "BCutAndroid"

if __name__ == '__main__':
    # full kwargs 上传测试
    # stream_gears.upload(
    #     video_path=["examples/test.mp4"],
    #     cookie_file="cookies.json",
    #     title="测试视频标题",
    #     tid=171,  # 电子竞技分区
    #     tag="测试,上传,演示",
    #     copyright=1,  # 自制
    #     source="",
    #     desc="这是一个测试视频的描述",
    #     dynamic="发布动态内容",
    #     cover="",  # 封面路径，空表示不设置封面
    #     dolby=0,  # 不开启杜比音效
    #     lossless_music=0,  # 不开启Hi-Res
    #     no_reprint=0,  # 允许转载
    #     charging_pay=0,  # 不开启充电
    #     up_close_reply=False,  # 不关闭评论
    #     up_selection_reply=False,  # 不开启精选评论
    #     up_close_danmu=False,  # 不关闭弹幕
    #     limit=3,  # 并发数
    #     desc_v2=[],  # 视频简介v2
    #     dtime=None,  # 立即发布
    #     line=stream_gears.UploadLine.Bda2,  # 指定上传线路
    #     extra_fields=None,  # 额外字段
    #     submit="BCutAndroid",  # 使用必剪安卓版接口提交
    #     proxy=None  # 不使用代理
    # )

    # 视频简介v2
    def creditsToDesc_v2(desc: str, credits: List[Dict[str, Any]]):
        desc_v2 = []
        desc_v2_tmp = desc
        for credit in credits:
            try:
                num = desc_v2_tmp.index("@credit")
                desc_v2.append({
                    "raw_text": " " + desc_v2_tmp[:num],
                    "biz_id": "",
                    "type": 1
                })
                desc_v2.append({
                    "raw_text": credit["username"],
                    "biz_id": str(credit["uid"]),
                    "type": 2
                })
                desc = desc.replace(
                    "@credit", "@" + credit["username"] + "  ", 1)
                desc_v2_tmp = desc_v2_tmp[num + 7:]
            except IndexError:
                print('简介中的@credit占位符少于credits的数量,替换失败')
        desc_v2.append({
            "raw_text": str(desc_v2_tmp),
            "biz_id": "",
            "type": 1
        })
        # desc_v2[0]["raw_text"] = desc_v2[0]["raw_text"][1:]  # 开头空格会导致识别简介过长
        return desc_v2

    # 额外字段（JSON格式）
    extra_fields_json = '{"is_self_only": "1"}'

    # 测试
    stream_gears.upload(
        video_path=video_path,
        cookie_file=cookie_file,
        copyright=1,
        extra_fields=extra_fields_json,
        submit="b-cut-android",
        title="必剪投稿接口测试"
    )

    print("所有测试用例已完成！")



DOUYIN_USER_AGENT = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/92.0.4515.159 Safari/537.36'

def load_webmssdk(js_file: str):
    with open(js_file, 'r', encoding='utf-8') as f:
        return f.read()

def get_signature(X_MS_STUB: str):
    try:
        import jsengine
        ctx = jsengine.jsengine()
        js_dom = f"""
document = {{}}
window = {{}}
navigator = {{
'userAgent': '{DOUYIN_USER_AGENT}'
}}
""".strip()
        js_enc = load_webmssdk('biliup\Danmaku\douyin_util\webmssdk.js')
        final_js = js_dom + js_enc
        ctx.eval(final_js)
        function_caller = f"get_sign('{X_MS_STUB}')"
        signature = ctx.eval(function_caller)
        print("signature: ", signature)
        return signature
    except:
        raise
    return "00000000"

get_signature("69a78110dbe05a916c750237d701907e")


import hashlib
import logging
import os

logger = logging.getLogger('biliup')

class DouyinDanmakuUtils:
    @staticmethod
    def load_webmssdk(js_file: str):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        js_path = os.path.join(dir_path, js_file)
        with open(js_path, 'r', encoding='utf-8') as f:
            return f.read()

    @staticmethod
    def get_user_unique_id() -> str:
        import random
        return str(random.randint(7300000000000000000, 7999999999999999999))

    @staticmethod
    def get_x_ms_stub(params: dict) -> str:
        sig_params = ','.join([f'{k}={v}' for k, v in params.items()])
        return hashlib.md5(sig_params.encode()).hexdigest()

    @staticmethod
    def get_signature(x_ms_stub: str):
        from biliup.plugins.douyin import DouyinUtils
        try:
            import jsengine
            ctx = jsengine.jsengine()
            js_dom = f"""
document = {{}}
window = {{}}
navigator = {{
  'userAgent': '{DouyinUtils.DOUYIN_USER_AGENT}'
}}
""".strip()
            js_enc = DouyinDanmakuUtils.load_webmssdk('webmssdk.js')
            final_js = js_dom + js_enc
            ctx.eval(final_js)
            function_caller = f"get_sign('{x_ms_stub}')"
            signature = ctx.eval(function_caller)
            # print("signature: ", signature)
            return signature
        except:
            logger.exception("get_signature error")
        return "00000000"


import logging
import sys

# logging.SafeRotatingFileHandler = SafeRotatingFileHandler
# log_file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'configlog.ini')
# logging.config.fileConfig(log_file_path)


def new_hook(t, v, tb):
    if issubclass(t, KeyboardInterrupt):
        sys.__excepthook__(t, v, tb)
        return
    logging.getLogger('biliup').error("Uncaught exception:", exc_info=(t, v, tb))


sys.excepthook = new_hook




import hashlib

import biliup.common.util
from biliup.config import config
from biliup.Danmaku import DanmakuClient
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase
from ..plugins import logger, match1

VALID_URL_BASE = r"https?://twitcasting\.tv/([^/]+)"


@Plugin.download(regexp=VALID_URL_BASE)
class Twitcasting(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)
        self.twitcasting_danmaku = config.get('twitcasting_danmaku', False)
        self.twitcasting_password = config.get('user.twitcasting_password')
        self.twitcasting_quality = config.get('twitcasting_quality')
        self.twitcasting_cookie = config.get('user', {}).get('twitcasting_cookie')
        self.fake_headers['referer'] = "https://twitcasting.tv/"

        # TODO 传递过于繁琐
        self._movie_id = None

    async def acheck_stream(self, is_check=False):
        cookies = {}
        if self.twitcasting_cookie:
            cookies = dict(item.strip().split("=", 1) for item in self.twitcasting_cookie.split(";"))
        if self.twitcasting_password:
            cookies["wpass"] = hashlib.md5(self.twitcasting_password.encode(encoding='UTF-8')).hexdigest()
        if cookies:
            self.fake_headers['cookie'] = "; ".join([f"{k}={v}" for k, v in cookies.items()])

        room_html = (await biliup.common.util.client.get(self.url, headers=self.fake_headers)).text
        if 'Enter the secret word to access' in room_html:
            logger.warning(f"{Twitcasting.__name__}: {self.url}: 直播间需要密码")
            return False

        # 尺寸不合适
        # self.live_cover_url = match1(room_html, r'<meta property="og:image" content="([^"]*)"')
        self.room_title = match1(room_html, r'<meta name="twitter:title" content="([^"]*)"')
        uploader_id = match1(room_html, r'<meta name="twitter:creator" content="([^"]*)"')

        response = await biliup.common.util.client.get(
            f'https://twitcasting.tv/streamserver.php?target={uploader_id}&mode=client&player=pc_web',
            headers=self.fake_headers)
        if response.status_code != 200:
            logger.warning(f"{Twitcasting.__name__}: {self.url}: 获取错误，本次跳过")
            return False
        stream_info = response.json()
        if not stream_info:
            logger.warning(f"{Twitcasting.__name__}: {self.url}: 直播间地址错误")
            return False
        if not stream_info['movie']['live']:
            logger.debug(f"{Twitcasting.__name__}: {self.url}: 未开播")
            return False

        self._movie_id = stream_info['movie']['id']

        if not stream_info.get("tc-hls", {}).get("streams"):
            logger.error(f"{Twitcasting.__name__}: {self.url}: 未获取到到直播流 => {stream_info}")
            return False

        stream_url = None
        quality_levels = ["high", "medium", "low"]
        streams = stream_info["tc-hls"]["streams"]
        if self.twitcasting_quality:
            quality_levels_filter = quality_levels[quality_levels.index(self.twitcasting_quality):]
            for quality_level in quality_levels_filter:
                if quality_level in streams:
                    stream_url = streams[quality_level]
                    break
        if not stream_url:
            for quality_level in quality_levels:
                if quality_level in streams:
                    stream_url = streams[quality_level]
                    break
        if not stream_url:
            stream_url = next(iter(streams.values()))

        if not stream_url:
            logger.error(f"{Twitcasting.__name__}: {self.url}: 未查找到直播流 => {stream_info}")
            return False

        self.raw_stream_url = stream_url

        return True

    def danmaku_init(self):
        if self.twitcasting_danmaku:
            self.danmaku = DanmakuClient(self.url, self.gen_download_filename(), {
                'movie_id': self._movie_id,
                'password': self.twitcasting_password,
            })

#
# class TwitcastingUtils:
#     import hashlib
#
#     def _getBroadcaster(html_text: str) -> dict:
#         _info = {}
#         _info['ID'] = match1(html_text, VALID_URL_BASE)
#         _info['Title'] = match1(
#             html_text,
#             r'<meta name="twitter:title" content="([^"]*)"'
#         )
#         _info['MovieID'] = match1(
#             match1(
#                 html_text,
#                 r'<meta name="twitter:image" content="([^"]*)"'
#             ),
#             r'/(\d+)'
#         )
#         _info['web-authorize-session-id'] = json.loads(
#             match1(
#                 html_text,
#                 r'<meta name="tc-page-variables" content="([^"]+)"'
#             ).replace(
#                 '&quot;',
#                 '"'
#             )
#         ).get('web-authorize-session-id')
#         return _info
#
#     def _generate_authorizekey(salt: str, timestamp: str, method: str, pathname: str, search: str,
#                                sessionid: str) -> str:
#         _hash_str = salt + timestamp + method + pathname + search + sessionid
#         return str(timestamp + "." + TwitcastingUtils.hashlib.sha256(_hash_str.encode()).hexdigest())
# '''
# X-Web-Authorizekey 可在 PlayerPage2.js 文件中
# 通过 return ""[u(413)](m, ".")[u(413)](f) 所在的方法计算而出
# 由 salt + 10位 timestamp + 接口Method大写 + 接口pathname + 接口search + web-authorize-session-id 拼接后
# 再经过 SHA-256 处理，最后在字符串前面拼接上 10位 timestamp 和 dot 得到
# '''
# __n = int(time.time() * 1000)
# _salt = "d6g97jormun44naq"
# _time = str(__n)[:10]
# _method = "GET"
# _pathname = f"/users/{boardcasterInfo['ID']}/latest-movie"
# _search = "?__n=" + str(__n)
#
# s.headers.update({
#     "X-Web-Authorizekey": TwitcastingUtils._generate_authorizekey(
#         _salt,
#         _time,
#         _method,
#         _pathname,
#         _search,
#         boardcasterInfo['web-authorize-session-id']
#     ),
#     "X-Web-Sessionid": boardcasterInfo['web-authorize-session-id'],
# })
# params = {"__n": __n}
# r = s.get(f"https://frontendapi.twitcasting.tv{_pathname}", params=params, timeout=5).json()
# if not r['movie']['is_on_live']:
#     return False
#
# if boardcasterInfo['ID']:
#     params = {
#         "mode": "client",
#         "target": boardcasterInfo['ID']
#     }
#     _stream_info = s.get("https://twitcasting.tv/streamserver.php", params=params, timeout=5).json()
#     if not _stream_info['movie']['live']:
#         return False
#     if is_check:
#         return True



import random
import re
import subprocess
import time

import biliup.common.util
from biliup.config import config
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase
from ..plugins import logger


@Plugin.download(regexp=r'(?:https?://)?(?:(?:www|m|live)\.)?nicovideo\.jp')
class Nico(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)

    async def acheck_stream(self, is_check=False):
        try:
            response = await biliup.common.util.client.get(self.url, timeout=5)
            # 正则表达式
            pattern = r'"name":"(.*?)","description":"(.*?)"'
            # 执行匹配
            matches = re.findall(pattern, response.text)[0]
            self.room_title = matches[0]
        except:
            logger.info("获取标题失败")
        port = random.randint(1025, 65535)
        stream_shell = [
            "streamlink",
            "--player-external-http",  # 为外部程序提供流媒体数据
            "--player-external-http-port", str(port),  # 对外部输出流的端口
            self.url, "best"  # 流链接
        ]
        if config.get('user', {}).get('niconico-email') is not None:
            niconico_email = "--niconico-email " + config.get('user', {}).get('niconico-email')
            stream_shell.insert(1, niconico_email)
        if config.get('user', {}).get('niconico-password') is not None:
            niconico_password = "--niconico-password " + config.get('user', {}).get('niconico-password')
            stream_shell.insert(1, niconico_password)
        if config.get('user', {}).get('niconico-user-session') is not None:
            niconico_user_session = "--niconico-user-session " + config.get('user', {}).get('niconico-user-session')
            stream_shell.insert(1, niconico_user_session)
        if config.get('user', {}).get('niconico-purge-credentials') is not None:
            niconico_purge_credentials = "--niconico-purge-credentials " + config.get('user', {}).get('niconico-purge-credentials')
            stream_shell.insert(1, niconico_purge_credentials)
        self.proc = subprocess.Popen(stream_shell)
        self.raw_stream_url = f"http://localhost:{port}"
        i = 0
        while i < 5:
            if not (self.proc.poll() is None):
                return
            time.sleep(1)
            i += 1
        return True

    def close(self):
        try:
            if self.proc is not None:
                self.proc.terminate()
        except:
            logger.exception(f'terminate {self.fname} failed')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# filename: __logger.py

# Tencent is pleased to support the open source community by making Tars available.
#
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
#

'''
@version: 0.01
@brief: 日志模块
'''

# 仅用于调试

import logging
import os
import re
from logging.handlers import RotatingFileHandler

tarsLogger = logging.getLogger('TARS client')
strToLoggingLevel = {
    "critical": logging.CRITICAL,
    "error": logging.ERROR,
    "warn": logging.WARNING,
    "info": logging.INFO,
    "debug": logging.DEBUG,
    "none": logging.NOTSET
}
#console = logging.StreamHandler()
# console.setLevel(logging.DEBUG)
#filelog = logging.FileHandler('tars.log')
# filelog.setLevel(logging.DEBUG)
#formatter = logging.Formatter('%(asctime)s | %(levelname)8s | [%(name)s] %(message)s', '%Y-%m-%d %H:%M:%S')
# console.setFormatter(formatter)
# filelog.setFormatter(formatter)
# tarsLogger.addHandler(console)
# tarsLogger.addHandler(filelog)
# tarsLogger.setLevel(logging.DEBUG)
# tarsLogger.setLevel(logging.INFO)
# tarsLogger.setLevel(logging.ERROR)


def createLogFile(filename):
    if filename.endswith('/'):
        raise ValueError("The logfile is a dir not a file")
    if os.path.exists(filename) and os.path.isfile(filename):
        pass
    else:
        fileComposition = str.split(filename, '/')
        print(fileComposition)
        currentFile = ''
        for item in fileComposition:
            if item == fileComposition[-1]:
                currentFile += item
                if not os.path.exists(currentFile) or not os.path.isfile(currentFile):
                    while True:
                        try:
                            os.mknod(currentFile)
                            break
                        except OSError as msg:
                            errno = re.findall(r"\d+", str(msg))
                            if len(errno) > 0 and errno[0] == '17':
                                currentFile += '.log'
                                continue
                break
            currentFile += (item + '/')
            if not os.path.exists(currentFile):
                os.mkdir(currentFile)


def initLog(logpath, logsize, lognum, loglevel):
    createLogFile(logpath)
    handler = RotatingFileHandler(filename=logpath, maxBytes=logsize,
                                  backupCount=lognum)
    formatter = logging.Formatter(
        '%(asctime)s | %(levelname)6s | [%(filename)18s:%(lineno)4d] | [%(thread)d] %(message)s', '%Y-%m-%d %H:%M:%S')
    handler.setFormatter(formatter)
    tarsLogger.addHandler(handler)
    if loglevel in strToLoggingLevel:
        tarsLogger.setLevel(strToLoggingLevel[loglevel])
    else:
        tarsLogger.setLevel(strToLoggingLevel["error"])


if __name__ == '__main__':
    tarsLogger.debug('debug log')
    tarsLogger.info('info log')
    tarsLogger.warning('warning log')
    tarsLogger.error('error log')
    tarsLogger.critical('critical log')



from .decorators import Plugin


def invert_dict(d: dict):
    inverse_dict = {}
    for k, v in d.items():
        if isinstance(v, list):
            for item in v:
                inverse_dict[item] = k
        else:
            inverse_dict[v] = k
    return inverse_dict


__all__ = ['invert_dict', 'Plugin']



# Tencent is pleased to support the open source community by making Tars available.
#
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
#


from .__util import util


class RequestPacket(util.struct):
    mapcls_context = util.mapclass(util.string, util.string)
    mapcls_status = util.mapclass(util.string, util.string)

    def __init__(self):
        self.iVersion = 0
        self.cPacketType = 0
        self.iMessageType = 0
        self.iRequestId = 0
        self.sServantName = ''
        self.sFuncName = ''
        self.sBuffer = bytes()
        self.iTimeout = 0
        self.context = RequestPacket.mapcls_context()
        self.status = RequestPacket.mapcls_status()

    @staticmethod
    def writeTo(oos, value):
        oos.write(util.int16, 1, value.iVersion)
        oos.write(util.int8, 2, value.cPacketType)
        oos.write(util.int32, 3, value.iMessageType)
        oos.write(util.int32, 4, value.iRequestId)
        oos.write(util.string, 5, value.sServantName)
        oos.write(util.string, 6, value.sFuncName)
        oos.write(util.bytes, 7, value.sBuffer)
        oos.write(util.int32, 8, value.iTimeout)
        oos.write(RequestPacket.mapcls_context, 9, value.context)
        oos.write(RequestPacket.mapcls_status, 10, value.status)

    @staticmethod
    def readFrom(ios):
        value = RequestPacket()
        value.iVersion = ios.read(util.int16, 1, True, 0)
        # print(("iVersion = %d" % value.iVersion))
        value.cPacketType = ios.read(util.int8, 2, True, 0)
        # print(("cPackerType = %d" % value.cPacketType))
        value.iMessageType = ios.read(util.int32, 3, True, 0)
        # print(("iMessageType = %d" % value.iMessageType))
        value.iRequestId = ios.read(util.int32, 4, True, 0)
        # print(("iRequestId = %d" % value.iRequestId))
        value.sServantName = ios.read(util.string, 5, True, '22222222')
        value.sFuncName = ios.read(util.string, 6, True, '')
        value.sBuffer = ios.read(util.bytes, 7, True, value.sBuffer)
        value.iTimeout = ios.read(util.int32, 8, True, 0)
        value.context = ios.read(
            RequestPacket.mapcls_context, 9, True, value.context)
        value.status = ios.read(
            RequestPacket.mapcls_status, 10, True, value.status)
        return value


class ResponsePacket(util.struct):
    __tars_class__ = "tars.RpcMessage.ResponsePacket"
    mapcls_status = util.mapclass(util.string, util.string)

    def __init__(self):
        self.iVersion = 0
        self.cPacketType = 0
        self.iRequestId = 0
        self.iMessageType = 0
        self.iRet = 0
        self.sBuffer = bytes()
        self.status = RequestPacket.mapcls_status()

    @staticmethod
    def writeTo(oos, value):
        oos.write(util.int16, 1, value.iVersion)
        oos.write(util.int8, 2, value.cPacketType)
        oos.write(util.int32, 3, value.iRequestId)
        oos.write(util.int32, 4, value.iMessageType)
        oos.write(util.int32, 5, value.iRet)
        oos.write(util.bytes, 6, value.sBuffer)
        oos.write(value.mapcls_status, 7, value.status)

    @staticmethod
    def readFrom(ios):
        value = ResponsePacket()
        value.iVersion = ios.read(util.int16, 1, True)
        value.cPacketType = ios.read(util.int8, 2, True)
        value.iRequestId = ios.read(util.int32, 3, True)
        value.iMessageType = ios.read(util.int32, 4, True)
        value.iRet = ios.read(util.int32, 5, True)
        value.sBuffer = ios.read(util.bytes, 6, True)
        value.status = ios.read(value.mapcls_status, 7, True)
        return value



import asyncio

import stream_gears

async def main():
    # from biliup.common.util import client, loop
    # print(loop)
    config = stream_gears.config_bindings()
    print(config)
    print(config.get("twitcasting_password", "313213"))
    print(config.get("file_size"))
    print(config.get("file_size"))
    print(config.get("streamers"))
    # await loop.run_in_executor(None, stream_gears.main_loop)
    # stream_gears.main_loop()


if __name__ == '__main__':
    asyncio.run(main())


from biliup.common.tars import tarscore
from .__util import auto_decode_fields
from ..wup_struct.UserId import HuyaUserId
from ..wup_struct.BeginLiveNotice import HuyaBeginLiveNotice
from ..wup_struct.StreamSettingNotice import HuyaStreamSettingNotice


class HuyaGetLivingInfoReq(tarscore.struct):

    __tars_class__ = "Huya.GetLivingInfoReq"

    def __init__(self):
        self.tId: HuyaUserId = HuyaUserId()
        self.lTopSid: tarscore.int64 = 0
        self.lSubSid: tarscore.int64 = 0
        self.lPresenterUid: tarscore.int64 = 0

    @staticmethod
    def writeTo(oos: tarscore.TarsOutputStream, value):
        oos.write(HuyaUserId, 0, value.tId)
        oos.write(tarscore.int64, 1, value.lTopSid)
        oos.write(tarscore.int64, 2, value.lSubSid)
        oos.write(tarscore.int64, 3, value.lPresenterUid)

    @staticmethod
    def readFrom(ios: tarscore.TarsInputStream):
        value = HuyaGetLivingInfoReq()
        value.tId = ios.read(HuyaUserId, 0, False)
        value.lTopSid = ios.read(tarscore.int64, 1, False)
        value.lSubSid = ios.read(tarscore.int64, 2, False)
        value.lPresenterUid = ios.read(tarscore.int64, 3, False)
        return value



class HuyaGetLivingInfoRsp(tarscore.struct):

    __tars_class__ = "Huya.GetLivingInfoRsp"

    def __init__(self):
        self.bIsLiving: tarscore.int32 = 0
        self.tNotice: HuyaBeginLiveNotice = HuyaBeginLiveNotice()
        self.tStreamSettingNotice: HuyaStreamSettingNotice = HuyaStreamSettingNotice()
        self.bIsSelfLiving: tarscore.int32 = 0

    @staticmethod
    def writeTo(oos: tarscore.TarsOutputStream, value):
        oos.write(tarscore.int32, 0, value.bIsLiving)
        oos.write(HuyaBeginLiveNotice, 1, value.tNotice)
        oos.write(HuyaStreamSettingNotice, 2, value.tStreamSettingNotice)
        oos.write(tarscore.int32, 3, value.bIsSelfLiving)

    @staticmethod
    def readFrom(ios: tarscore.TarsInputStream):
        value = HuyaGetLivingInfoRsp()
        value.bIsLiving = ios.read(tarscore.int32, 0, False)
        value.tNotice = ios.read(HuyaBeginLiveNotice, 1, False)
        value.tStreamSettingNotice = ios.read(HuyaStreamSettingNotice, 2, False)
        value.bIsSelfLiving = ios.read(tarscore.int32, 3, False)
        return value



from biliup.common.tars import tarscore
from .__util import auto_decode_fields

@auto_decode_fields
class HuyaGetCdnTokenReq(tarscore.struct):

    __tars_class__ = "HUYA.GetCdnTokenReq"

    def __init__(self):
        self.url: tarscore.string = ""
        self.cdnType: tarscore.string = ""
        self.streamName: tarscore.string = ""
        self.presenterUid: tarscore.int32 = 0

    @staticmethod
    def writeTo(oos: tarscore.TarsOutputStream, value):
        oos.write(tarscore.string, 0, value.url)
        oos.write(tarscore.string, 1, value.cdnType)
        oos.write(tarscore.string, 2, value.streamName)
        oos.write(tarscore.int64, 3, value.presenterUid)

    @staticmethod
    def readFrom(ios: tarscore.TarsInputStream):
        value = HuyaGetCdnTokenReq()
        value.url = ios.read(tarscore.string, 0, False)
        value.cdnType = ios.read(tarscore.string, 1, False)
        value.streamName = ios.read(tarscore.string, 2, False)
        value.presenterUid = ios.read(tarscore.int64, 3, False)
        return value

    def as_dict(self):
        return self.__dict__.copy()

@auto_decode_fields
class HuyaGetCdnTokenRsp(tarscore.struct):

    __tars_class__ = "HUYA.GetCdnTokenRsp"

    def __init__(self):
        self.url: tarscore.string = ""
        self.cdnType: tarscore.string = ""
        self.streamName: tarscore.string = ""
        self.presenterUid: tarscore.int64 = 0
        self.antiCode: tarscore.string = ""
        self.sTime: tarscore.string = ""
        self.flvAntiCode: tarscore.string = ""
        self.hlsAntiCode: tarscore.string = ""

    @staticmethod
    def writeTo(oos: tarscore.TarsOutputStream, value):
        oos.write(tarscore.string, 0, value.url)
        oos.write(tarscore.string, 1, value.cdnType)
        oos.write(tarscore.string, 2, value.streamName)
        oos.write(tarscore.int64, 3, value.presenterUid)
        oos.write(tarscore.string, 4, value.antiCode)
        oos.write(tarscore.string, 5, value.sTime)
        oos.write(tarscore.string, 6, value.flvAntiCode)
        oos.write(tarscore.string, 7, value.hlsAntiCode)

    @staticmethod
    def readFrom(ios: tarscore.TarsInputStream):
        value = HuyaGetCdnTokenRsp()
        value.url = ios.read(tarscore.string, 0, False)
        value.cdnType = ios.read(tarscore.string, 1, False)
        value.streamName = ios.read(tarscore.string, 2, False)
        value.presenterUid = ios.read(tarscore.int64, 3, False)
        value.antiCode = ios.read(tarscore.string, 4, False)
        value.sTime = ios.read(tarscore.string, 5, False)
        value.flvAntiCode = ios.read(tarscore.string, 6, False)
        value.hlsAntiCode = ios.read(tarscore.string, 7, False)
        return value

    def as_dict(self):
        return self.__dict__.copy()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import os
import queue
import threading
import subprocess
import time
from urllib.parse import urlparse


logger = logging.getLogger('biliup.engine.sync_downloader')


def pad_file_to_size(filename, min_file_size):
    """
    若文件大小不足 min_file_size，则在文件末尾填充 0x00 至目标大小。
    """
    if not os.path.exists(filename):
        return

    current_size = os.path.getsize(filename)
    if current_size < min_file_size:
        need_pad = min_file_size - current_size
        print(f"[pad_file_to_size] 补齐文件 {filename}："
              f"填充 {need_pad} 字节 0x00 使其达到 {min_file_size} 字节")
        with open(filename, "ab") as f:
            f.write(b"\x00" * need_pad)


class SyncDownloader:
    """
    同步下载-切片类
    说明：
      1. 在 run() 方法中，单线程循环执行录制逻辑；
      2. 每次启动一段 ffmpeg 录制，并读取 streamlink stdout 作为输入；
      3. 当 ffmpeg 录制结束后，杀掉当前的 streamlink 进程并补齐文件大小；
      4. 若中途发现 streamlink 无数据可读（EOF），则说明没有更多内容可下载，结束整个程序。
    """

    def __init__(self,
                 stream_url="http://localhost:8888/stream0/index.m3u8",
                 headers={"a": "1"},
                 segment_duration=10,
                 max_file_size=100,
                 output_prefix="segment_",
                 video_queue=None):
        """
        :param stream_url:   拉流地址

        :param segment_duration: 每段录制时长（秒）（暂时不用）
        :param max_file_size:     文件最小大小，不足时进行 0x00 填充 单位 MB
        :param read_block_size:   从 streamlink stdout 读取数据时的单次块大小
        :param output_prefix:     输出文件名前缀
        """
        self.stream_url = stream_url
        self.quality = "best"
        self.headers = headers
        self.segment_duration = segment_duration
        self.read_block_size = 500
        self.max_file_size = max_file_size
        self.output_prefix = output_prefix

        self.video_queue: queue.SimpleQueue = video_queue
        self.stop_event = threading.Event()

    def run_ffmpeg_with_url(self, ffmpeg_cmd, output_filename):
        with subprocess.Popen(ffmpeg_cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE) as ffmpeg_proc:
            logger.info("[run] 启动 ffmpeg...")
            if output_filename == "-":
                data = ffmpeg_proc.stdout.read(self.read_block_size)  # 读取第一个 data
                if not data:  # 如果第一个 data 为空
                    logger.info("[run] ffmpeg 没有输出数据，返回 False")
                    err = ffmpeg_proc.stderr.read()
                    if err:
                        logger.error("[run] ffmpeg err " + err.decode("utf-8", errors="replace"))
                    return False
                self.video_queue.put(data)  # 将第一个数据放入队列
                while True:
                    data = ffmpeg_proc.stdout.read(self.read_block_size)
                    if not data:
                        logger.info("[run] ffmpeg stdout 已到达 EOF。结束本段写入。")
                        break
                    self.video_queue.put(data)
            ffmpeg_proc.wait()
            # 输出 ffmpeg 的错误信息（如果有的话）
            # err = ffmpeg_proc.stderr.read()
            # if err:
            #     logger.error("[run] ffmpeg err " + err.decode("utf-8", errors="replace"))
        return True  # 如果正常执行，返回 True

    def run_streamlink_with_ffmpeg(self, streamlink_cmd, ffmpeg_cmd, output_filename):
        with subprocess.Popen(streamlink_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as streamlink_proc:
            logger.info("[run] 启动 streamlink...")
            with subprocess.Popen(ffmpeg_cmd, stdin=streamlink_proc.stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as ffmpeg_proc:
                logger.info("[run] 启动 ffmpeg...")
                if output_filename == "-":
                    logger.info("[run] 读取 ffmpeg stdout...")
                    # 读取第一个数据
                    data = ffmpeg_proc.stdout.read(self.read_block_size)
                    if not data:  # 如果第一个数据为空
                        logger.info("[run] ffmpeg 没有输出数据，返回 False")
                        streamlink_proc.kill()  # 终止 streamlink 进程
                        ffmpeg_proc.kill()  # 终止 ffmpeg 进程
                        # 打印错误输出
                        ffmpeg_err = ffmpeg_proc.stderr.read()
                        if ffmpeg_err:
                            logger.error("[run] ffmpeg err " + ffmpeg_err.decode("utf-8", errors="replace"))
                        streamlink_err = streamlink_proc.stderr.read()
                        if streamlink_err:
                            logger.error("[run] streamlink err " + streamlink_err.decode("utf-8", errors="replace"))
                        return False

                    self.video_queue.put(data)  # 将第一个数据放入队列
                    # 继续读取剩余数据
                    while True:
                        data = ffmpeg_proc.stdout.read(self.read_block_size)
                        if not data:
                            logger.info("[run] ffmpeg stdout 已到达 EOF。结束本段写入。")
                            break
                        self.video_queue.put(data)

                ffmpeg_proc.wait()
                logger.info("[run] ffmpeg 已到达输出大小并退出。结束本段写入。")
                # 打印 ffmpeg 子进程的错误输出
                # ffmpeg_err = ffmpeg_proc.stderr.read()
                # if ffmpeg_err:
                # logger.error("[run] ffmpeg err " + ffmpeg_err.decode("utf-8", errors="replace"))
            # 打印 streamlink 子进程的错误输出
            # streamlink_err = streamlink_proc.stderr.read()
            # if streamlink_err:
            #     logger.error("[run] streamlink err " + streamlink_err.decode("utf-8", errors="replace"))
        return True  # 如果一切正常，返回 True

    def build_ffmpeg_cmd(self, input_source, output_filename, headers, segment_duration):
        cmd = [
            "ffmpeg",
            "-loglevel", "error",
            # "-"  # 覆盖输出文件
        ]
        if headers:
            cmd += ["-headers", ''.join(f'{key}: {value}\r\n' for key, value in headers.items())]
        for i in [
            "-fflags", "+genpts",
            "-i", input_source,  # 输入源
            # "-t", str(segment_duration),
            "-fs", f"{self.max_file_size}M",
            "-c:v", "copy",
            "-c:a", "copy",
            "-reset_timestamps", "1",
            "-avoid_negative_ts", "1",
            "-movflags", "+frag_keyframe+empty_moov",
            "-f", "matroska",
            "-",
        ]:
            cmd.append(i)
        return cmd

    def run(self):
        """
        主逻辑：循环进行分段录制。
        - 每段录制：
          1) 启动 streamlink；若 EOF 则退出。
          2) 启动 ffmpeg (带 -fs 参数限制输出大小)；
          3) 从 streamlink stdout 读数据，写给 ffmpeg stdin；
          4) ffmpeg 到时长后退出，然后杀掉 streamlink；
          5) 进入下一段，如此往复。
        - 若中途发现 streamlink 无数据（EOF）则跳出循环。
        """

        file_index = 1
        retry_count = 0
        while True:
            if self.stop_event.is_set():
                break
            if retry_count >= 5:
                logger.info("这个直播流已经失效，停止下载器")
                return

            output_filename = f"{self.output_prefix}{file_index:03d}.mkv"
            # print(f"\n[run] ========== 准备录制第 {file_index} 段：{output_filename} ==========")
            # logging.info(f"\n[run] == 当前下载流地址：{self.stream_url} ==")
            logger.info(f"\n[run] == 准备录制第 {file_index} 段：{output_filename} ==")
            output_filename = "-"
            is_hls = '.m3u8' in urlparse(self.stream_url).path
            if not is_hls:
                # print("[run] 输入源不是 HLS 地址，将直接使用 ffmpeg 进行录制。", self.stream_url)
                logger.info("[run] 输入源不是 HLS 地址，将直接使用 ffmpeg 进行录制。")
                ffmpeg_cmd = self.build_ffmpeg_cmd(self.stream_url, output_filename,
                                                   self.headers, self.segment_duration)
                if not self.run_ffmpeg_with_url(ffmpeg_cmd, output_filename):
                    retry_count += 1
                    time.sleep(1)
                    continue
            else:
                # print("[run] 输入源是 HLS 地址，将使用 streamlink + ffmpeg 进行录制。")
                logger.info("[run] 输入源是 HLS 地址，将使用 streamlink + ffmpeg 进行录制。")
                if self.headers:
                    headers = []
                    for key, value in self.headers.items():
                        headers.extend(['--http-header', f'{key}={value}'])
                streamlink_cmd = [
                    'streamlink',
                    '--stream-segment-threads', '3',
                    '--hls-playlist-reload-attempts', '1',
                    *headers,
                    self.stream_url,
                    self.quality,
                    '-O'
                ]
                logger.info(f"[run] streamlink_cmd: {streamlink_cmd}")
                # output_filename = "-"
                ffmpeg_cmd = self.build_ffmpeg_cmd("pipe:0", output_filename, None, self.segment_duration)
                if not self.run_streamlink_with_ffmpeg(streamlink_cmd, ffmpeg_cmd, output_filename):
                    retry_count += 1
                    time.sleep(1)
                    continue

            # 6. 进入下一段
            # if file_index != 1:
            self.video_queue.put(None)  # 通知消费者线程本段录制结束
            file_index += 1


def main():
    slicer = SyncDownloader(
        stream_url="http://127.0.0.1:8888/live/index.m3u8",
        segment_duration=10,
        read_block_size=4096,
        output_prefix="segment_"
    )

    # ====【可选】消费者线程示例，演示如何拿到 video_queue 的数据====
    def consumer():
        file_index = 1
        while True:
            data_count = 0
            with open(f"output_{file_index}.mkv", "wb") as f:
                while True:
                    data = slicer.video_queue.get()  # 阻塞式获取
                    if data is None:
                        break
                    f.write(data)
                    data_count += len(data)
                    # print(f"[consumer] 写入文件 output_{file_index}.mkv，大小：{data_count} 字")
            print(f"[consumer] 写入文件 output_{file_index}.mkv，大小：{data_count} 字")

            if data_count < 100:
                print(f"[consumer] 无效文件，删除 output_{file_index}.mkv")
                os.remove(f"output_{file_index}.mkv")
                slicer.stop_event.set()
                break

            pad_file_to_size(f"output_{file_index}.mkv", 100 * 1024 * 1024)  # 补齐文件大小
            file_index += 1
            # if slicer.stop_event.is_set():
            # break
            # 在这里可以对 data 做进一步处理，比如再推到别的地方

    # 启动消费者线程
    t = threading.Thread(target=consumer, daemon=True)
    t.start()

    # 启动下载录制主逻辑
    slicer.run()

    # 停止消费者（如录制完毕后可执行）
    # slicer.video_queue.put(None)

    # t.join()


if __name__ == "__main__":
    main()



import asyncio
import logging
import subprocess
import sys
import os

logger = logging.getLogger('biliup')

global global_reloader


def has_extension(fname_list, *extension):
    for fname in fname_list:
        result = list(map(fname.endswith, extension))
        if True in result:
            return True
    return False


def is_docker():
    path = '/proc/self/cgroup'
    return (
            os.path.exists('/.dockerenv') or
            os.path.isfile(path) and any('docker' in line for line in open(path))
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# filename: __adapterproxymanager.py_compiler

# Tencent is pleased to support the open source community by making Tars available.
#
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
#

'''
@version: 0.01
@brief: 将rpc部分中的adapterproxymanager抽离出来，实现不同的负载均衡
'''

import os
import random
import select
import socket
import time
from enum import Enum

from . import exception
# 因为循环import的问题只能放这里，不能放文件开始处
from .QueryF import QueryFProxy
from .QueryF import QueryFPrxCallback
from .__TimeoutQueue import ReqMessage
from .__logger import tarsLogger
from .__trans import EndPointInfo
from .__trans import TcpTransceiver
from .__util import (LockGuard, NewLock, ConsistentHashNew)
from .exception import TarsException


class AdapterProxy:
    '''
    @brief: 每一个Adapter管理一个服务端端口的连接，数据收发
    '''

    def __init__(self):
        tarsLogger.debug('AdapterProxy:__init__')
        self.__closeTrans = False
        self.__trans = None
        self.__object = None
        self.__reactor = None
        self.__lock = None
        self.__asyncProc = None
        self.__activeStateInReg = True

    @property
    def activatestateinreg(self):
        return self.__activeStateInReg

    @activatestateinreg.setter
    def activatestateinreg(self, value):
        self.__activeStateInReg = value

    def __del__(self):
        tarsLogger.debug('AdapterProxy:__del__')

    def initialize(self, endPointInfo, objectProxy, reactor, asyncProc):
        '''
        @brief: 初始化
        @param endPointInfo: 连接对端信息
        @type endPointInfo: EndPointInfo
        @type objectProxy: ObjectProxy
        @type reactor: FDReactor
        @type asyncProc: AsyncProcThread
        '''
        tarsLogger.debug('AdapterProxy:initialize')
        self.__closeTrans = False
        self.__trans = TcpTransceiver(endPointInfo)
        self.__object = objectProxy
        self.__reactor = reactor
        # self.__lock = threading.Lock()
        self.__lock = NewLock()
        self.__asyncProc = asyncProc

    def terminate(self):
        '''
        @brief: 关闭
        '''
        tarsLogger.debug('AdapterProxy:terminate')
        self.setCloseTrans(True)

    def trans(self):
        '''
        @brief: 获取传输类
        @return: 负责网络传输的trans
        @rtype: Transceiver
        '''
        return self.__trans

    def invoke(self, reqmsg):
        '''
        @brief: 远程过程调用处理方法
        @param reqmsg: 请求响应报文
        @type reqmsg: ReqMessage
        @return: 错误码：0表示成功，-1表示连接失败
        @rtype: int
        '''
        tarsLogger.debug('AdapterProxy:invoke')
        assert(self.__trans)

        if (not self.__trans.hasConnected() and
                not self.__trans.isConnecting):
            # -1表示连接失败
            return -1

        reqmsg.request.iRequestId = self.__object.getTimeoutQueue().generateId()
        self.__object.getTimeoutQueue().push(reqmsg, reqmsg.request.iRequestId)

        self.__reactor.notify(self)

        return 0

    def finished(self, rsp):
        '''
        @brief: 远程过程调用返回处理
        @param rsp: 响应报文
        @type rsp: ResponsePacket
        @return: 函数是否执行成功
        @rtype: bool
        '''
        tarsLogger.debug('AdapterProxy:finished')
        reqmsg = self.__object.getTimeoutQueue().pop(rsp.iRequestId)
        if not reqmsg:
            tarsLogger.error(
                'finished, can not get ReqMessage, may be timeout, id: %d',
                rsp.iRequestId)
            return False

        reqmsg.response = rsp
        if reqmsg.type == ReqMessage.SYNC_CALL:
            return reqmsg.servant._finished(reqmsg)
        elif reqmsg.callback:
            self.__asyncProc.put(reqmsg)
            return True

        tarsLogger.error('finished, adapter proxy finish fail, id: %d, ret: %d',
                         rsp.iRequestId, rsp.iRet)
        return False

    # 检测连接是否失败，失效时重连
    def checkActive(self, forceConnect=False):
        '''
        @brief: 检测连接是否失效
        @param forceConnect: 是否强制发起连接，为true时不对状态进行判断就发起连接
        @type forceConnect: bool
        @return: 连接是否有效
        @rtype: bool
        '''
        tarsLogger.debug('AdapterProxy:checkActive')
        # self.__lock.acquire()
        lock = LockGuard(self.__lock)
        tarsLogger.info('checkActive, %s, forceConnect: %s',
                        self.__trans.getEndPointInfo(), forceConnect)

        if not self.__trans.isConnecting() and not self.__trans.hasConnected():
            self.doReconnect()

        # self.__lock.release()
        return self.__trans.isConnecting() or self.__trans.hasConnected()

    def doReconnect(self):
        '''
        @brief: 重新发起连接
        @return: None
        @rtype: None
        '''
        tarsLogger.debug('AdapterProxy:doReconnect')
        assert(self.__trans)

        self.__trans.reInit()
        tarsLogger.info('doReconnect, connect: %s, fd:%d',
                        self.__trans.getEndPointInfo(),
                        self.__trans.getFd())

        self.__reactor.registerAdapter(self, select.EPOLLIN | select.EPOLLOUT)

    def sendRequest(self):
        '''
        @brief: 把队列中的请求放到Transceiver的发送缓存里
        @return: 放入缓存的数据长度
        @rtype: int
        '''
        tarsLogger.debug('AdapterProxy:sendRequest')
        if not self.__trans.hasConnected():
            return False

        reqmsg = self.__object.popRequest()
        blen = 0
        while reqmsg:
            reqmsg.adapter = self
            buf = reqmsg.packReq()
            self.__trans.writeToSendBuf(buf)
            tarsLogger.info('sendRequest, id: %d, len: %d',
                            reqmsg.request.iRequestId, len(buf))
            blen += len(buf)
            # 合并一次发送的包 最大合并至8k 提高异步时客户端效率?
            if (self.__trans.getEndPointInfo().getConnType() == EndPointInfo.SOCK_UDP
                    or blen > 8192):
                break
            reqmsg = self.__object.popRequest()

        return blen

    def finishConnect(self):
        '''
        @brief: 使用的非阻塞socket连接不能立刻判断是否连接成功，
                在epoll响应后调用此函数处理connect结束后的操作
        @return: 是否连接成功
        @rtype: bool
        '''
        tarsLogger.debug('AdapterProxy:finishConnect')
        success = True
        errmsg = ''
        try:
            ret = self.__trans.getSock().getsockopt(
                socket.SOL_SOCKET, socket.SO_ERROR)
            if ret:
                success = False
                errmsg = os.strerror(ret)
        except Exception as msg:
            errmsg = msg
            success = False

        if not success:
            self.__reactor.unregisterAdapter(
                self, socket.EPOLLIN | socket.EPOLLOUT)
            self.__trans.close()
            self.__trans.setConnFailed()
            tarsLogger.error(
                'AdapterProxy finishConnect, exception: %s, error: %s',
                self.__trans.getEndPointInfo(), errmsg)
            return False
        self.__trans.setConnected()
        self.__reactor.notify(self)
        tarsLogger.info('AdapterProxy finishConnect, connect %s success',
                        self.__trans.getEndPointInfo())
        return True

    def finishInvoke(self, isTimeout):
        pass

    # 弹出请求报文
    def popRequest(self):
        pass

    def shouldCloseTrans(self):
        '''
        @brief: 是否设置关闭连接
        @return: 关闭连接的flag的值
        @rtype: bool
        '''
        return self.__closeTrans

    def setCloseTrans(self, closeTrans):
        '''
        @brief: 设置关闭连接flag的值
        @param closeTrans: 是否关闭连接
        @type closeTrans: bool
        @return: None
        @rtype: None
        '''
        self.__closeTrans = closeTrans


class QueryRegisterCallback(QueryFPrxCallback):
    def __init__(self, adpManager):
        self.__adpManager = adpManager
        super(QueryRegisterCallback, self).__init__()
        # QueryFPrxCallback.__init__(self)

    def callback_findObjectById4All(self, ret, activeEp, inactiveEp):
        eplist = [EndPointInfo(x.host, x.port, x.timeout, x.weight, x.weightType)
                  for x in activeEp if ret == 0 and x.istcp]
        ieplist = [EndPointInfo(x.host, x.port, x.timeout, x.weight, x.weightType)
                   for x in inactiveEp if ret == 0 and x.istcp]
        self.__adpManager.setEndpoints(eplist, ieplist)

    def callback_findObjectById4All_exception(self, ret):
        tarsLogger.error('callback_findObjectById4All_exception ret: %d', ret)


class EndpointWeightType(Enum):
    E_LOOP = 0
    E_STATIC_WEIGHT = 1


class AdapterProxyManager:
    '''
    @brief: 管理Adapter
    '''

    def __init__(self):
        tarsLogger.debug('AdapterProxyManager:__init__')
        self.__comm = None
        self.__object = None
        # __adps的key=str(EndPointInfo) value=[EndPointInfo, AdapterProxy, cnt]
        # cnt是访问次数
        self.__adps = {}
        self.__iadps = {}
        self.__newLock = None
        self.__isDirectProxy = True
        self.__lastFreshTime = 0
        self.__queryRegisterCallback = QueryRegisterCallback(self)
        self.__regAdapterProxyDict = {}
        self.__lastConHashPrxList = []
        self.__consistentHashWeight = None
        self.__weightType = EndpointWeightType.E_LOOP
        self.__update = True
        self.__lastWeightedProxyData = {}

    def initialize(self, comm, objectProxy, eplist):
        '''
        @brief: 初始化
        '''
        tarsLogger.debug('AdapterProxyManager:initialize')
        self.__comm = comm
        self.__object = objectProxy
        self.__newLock = NewLock()

        self.__isDirectProxy = len(eplist) > 0
        if self.__isDirectProxy:
            self.setEndpoints(eplist, {})
        else:
            self.refreshEndpoints()

    def terminate(self):
        '''
        @brief: 释放资源
        '''
        tarsLogger.debug('AdapterProxyManager:terminate')
        # self.__lock.acquire()
        lock = LockGuard(self.__newLock)
        for ep, epinfo in self.__adps.items():
            epinfo[1].terminate()
        self.__adps = {}
        self.__lock.release()

    def refreshEndpoints(self):
        '''
        @brief: 刷新服务器列表
        @return: 新的服务列表
        @rtype: EndPointInfo列表
        '''
        tarsLogger.debug('AdapterProxyManager:refreshEndpoints')
        if self.__isDirectProxy:
            return

        interval = self.__comm.getProperty(
            'refresh-endpoint-interval', float) / 1000
        locator = self.__comm.getProperty('locator')

        if '@' not in locator:
            raise exception.TarsRegistryException(
                'locator is not valid: ' + locator)

        now = time.time()
        last = self.__lastFreshTime
        epSize = len(self.__adps)
        if last + interval < now or (epSize <= 0 and last + 2 < now):
            queryFPrx = self.__comm.stringToProxy(QueryFProxy, locator)
            # 首次访问是同步调用，之后访问是异步调用
            if epSize == 0 or last == 0:
                ret, activeEps, inactiveEps = (
                    queryFPrx.findObjectById4All(self.__object.name()))
                # 目前只支持TCP
                eplist = [EndPointInfo(x.host, x.port, x.timeout, x.weight, x.weightType)
                          for x in activeEps if ret == 0 and x.istcp]
                ieplist = [EndPointInfo(x.host, x.port, x.timeout, x.weight, x.weightType)
                           for x in inactiveEps if ret == 0 and x.istcp]
                self.setEndpoints(eplist, ieplist)
            else:
                queryFPrx.async_findObjectById4All(self.__queryRegisterCallback,
                                                   self.__object.name())
            self.__lastFreshTime = now

    def getEndpoints(self):
        '''
        @brief: 获取可用服务列表 如果启用分组,只返回同分组的服务端ip
        @return: 获取节点列表
        @rtype: EndPointInfo列表
        '''
        tarsLogger.debug('AdapterProxyManager:getEndpoints')
        # self.__lock.acquire()
        lock = LockGuard(self.__newLock)
        ret = [x[1][0] for x in list(self.__adps.items())]
        # self.__lock.release()

        return ret

    def setEndpoints(self, eplist, ieplist):
        '''
        @brief: 设置服务端信息
        @para eplist: 活跃的被调节点列表
        @para ieplist: 不活跃的被调节点列表
        '''
        tarsLogger.debug('AdapterProxyManager:setEndpoints')
        adps = {}
        iadps = {}
        comm = self.__comm
        isNeedNotify = False
        # self.__lock.acquire()
        lock = LockGuard(self.__newLock)
        isStartStatic = True

        for ep in eplist:
            if ep.getWeightType() == 0:
                isStartStatic = False
            epstr = str(ep)
            if epstr in self.__adps:
                adps[epstr] = self.__adps[epstr]
                continue
            isNeedNotify = True
            self.__update = True
            adapter = AdapterProxy()
            adapter.initialize(ep, self.__object,
                               comm.getReactor(), comm.getAsyncProc())
            adapter.activatestateinreg = True
            adps[epstr] = [ep, adapter, 0]
        self.__adps, adps = adps, self.__adps

        for iep in ieplist:
            iepstr = str(iep)
            if iepstr in self.__iadps:
                iadps[iepstr] = self.__iadps[iepstr]
                continue
            isNeedNotify = True
            adapter = AdapterProxy()
            adapter.initialize(iep, self.__object,
                               comm.getReactor(), comm.getAsyncProc())
            adapter.activatestateinreg = False
            iadps[iepstr] = [iep, adapter, 0]
        self.__iadps, iadps = iadps, self.__iadps

        if isStartStatic:
            self.__weightType = EndpointWeightType.E_STATIC_WEIGHT
        else:
            self.__weightType = EndpointWeightType.E_LOOP

        # self.__lock.release()
        if isNeedNotify:
            self.__notifyEndpoints(self.__adps, self.__iadps)
        # 关闭已经失效的连接
        for ep in adps:
            if ep not in self.__adps:
                adps[ep][1].terminate()

    def __notifyEndpoints(self, actives, inactives):
        # self.__lock.acquire()
        lock = LockGuard(self.__newLock)
        self.__regAdapterProxyDict.clear()
        self.__regAdapterProxyDict.update(actives)
        self.__regAdapterProxyDict.update(inactives)
        # self.__lock.release()

    def __getNextValidProxy(self):
        '''
        @brief: 刷新本地缓存列表，如果服务下线了，要求删除本地缓存
        @return:
        @rtype: EndPointInfo列表
        @todo: 优化负载均衡算法
        '''
        tarsLogger.debug('AdapterProxyManager:getNextValidProxy')
        lock = LockGuard(self.__newLock)
        if len(self.__adps) == 0:
            raise TarsException("the activate adapter proxy is empty")

        sortedActivateAdp = sorted(
            list(self.__adps.items()), key=lambda item: item[1][2])
        # self.refreshEndpoints()
        # self.__lock.acquire()
        sortedActivateAdpSize = len(sortedActivateAdp)

        while sortedActivateAdpSize != 0:
            if sortedActivateAdp[0][1][1].checkActive():
                self.__adps[sortedActivateAdp[0][0]][2] += 1
                # 返回的是 adapterProxy
                return self.__adps[sortedActivateAdp[0][0]][1]
            sortedActivateAdp.pop(0)
            sortedActivateAdpSize -= 1
        # 随机重连一个可用节点
        adpPrx = list(self.__adps.items())[random.randint(
            0, len(self.__adps))][1][1]
        adpPrx.checkActive()
        return None
        # self.__lock.release()

    def __getHashProxy(self, reqmsg):
        if self.__weightType == EndpointWeightType.E_LOOP:
            if reqmsg.isConHash:
                return self.__getConHashProxyForNormal(reqmsg.hashCode)
            else:
                return self.__getHashProxyForNormal(reqmsg.hashCode)
        else:
            if reqmsg.isConHash:
                return self.__getConHashProxyForWeight(reqmsg.hashCode)
            else:
                return self.__getHashProxyForWeight(reqmsg.hashCode)

    def __getHashProxyForNormal(self, hashCode):
        tarsLogger.debug('AdapterProxyManager:getHashProxyForNormal')
        # self.__lock.acquire()
        lock = LockGuard(self.__newLock)
        regAdapterProxyList = sorted(
            list(self.__regAdapterProxyDict.items()), key=lambda item: item[0])

        allPrxSize = len(regAdapterProxyList)
        if allPrxSize == 0:
            raise TarsException("the adapter proxy is empty")
        hashNum = hashCode % allPrxSize

        if regAdapterProxyList[hashNum][1][1].activatestateinreg and regAdapterProxyList[hashNum][1][1].checkActive():
            epstr = regAdapterProxyList[hashNum][0]
            self.__regAdapterProxyDict[epstr][2] += 1
            if epstr in self.__adps:
                self.__adps[epstr][2] += 1
            elif epstr in self.__iadps:
                self.__iadps[epstr][2] += 1
            return self.__regAdapterProxyDict[epstr][1]
        else:
            if len(self.__adps) == 0:
                raise TarsException("the activate adapter proxy is empty")
            activeProxyList = list(self.__adps.items())
            actPrxSize = len(activeProxyList)
            while actPrxSize != 0:
                hashNum = hashCode % actPrxSize
                if activeProxyList[hashNum][1][1].checkActive():
                    self.__adps[activeProxyList[hashNum][0]][2] += 1
                    return self.__adps[activeProxyList[hashNum][0]][1]
                activeProxyList.pop(hashNum)
                actPrxSize -= 1
            # 随机重连一个可用节点
            adpPrx = list(self.__adps.items())[random.randint(
                0, len(self.__adps))][1][1]
            adpPrx.checkActive()
            return None

    def __getConHashProxyForNormal(self, hashCode):
        tarsLogger.debug('AdapterProxyManager:getConHashProxyForNormal')
        lock = LockGuard(self.__newLock)
        if len(self.__regAdapterProxyDict) == 0:
            raise TarsException("the adapter proxy is empty")
        if self.__consistentHashWeight is None or self.__checkConHashChange(self.__lastConHashPrxList):
            self.__updateConHashProxyWeighted()

        if len(self.__consistentHashWeight.nodes) > 0:
            conHashIndex = self.__consistentHashWeight.getNode(hashCode)
            if conHashIndex in self.__regAdapterProxyDict and self.__regAdapterProxyDict[conHashIndex][1].activatestateinreg and self.__regAdapterProxyDict[conHashIndex][1].checkActive():
                self.__regAdapterProxyDict[conHashIndex][2] += 1
                if conHashIndex in self.__adps:
                    self.__adps[conHashIndex][2] += 1
                elif conHashIndex in self.__iadps:
                    self.__iadps[conHashIndex][2] += 1
                return self.__regAdapterProxyDict[conHashIndex][1]
            else:
                if len(self.__adps) == 0:
                    raise TarsException("the activate adapter proxy is empty")
                activeProxyList = list(self.__adps.items())
                actPrxSize = len(activeProxyList)
                while actPrxSize != 0:
                    hashNum = hashCode % actPrxSize
                    if activeProxyList[hashNum][1][1].checkActive():
                        self.__adps[activeProxyList[hashNum][0]][2] += 1
                        return self.__adps[activeProxyList[hashNum][0]][1]
                    activeProxyList.pop(hashNum)
                    actPrxSize -= 1
                # 随机重连一个可用节点
                adpPrx = list(self.__adps.items())[random.randint(
                    0, len(self.__adps))][1][1]
                adpPrx.checkActive()
                return None
            pass
        else:
            return self.__getHashProxyForNormal(hashCode)

    def __getHashProxyForWeight(self, hashCode):
        return None
        pass

    def __getConHashProxyForWeight(self, hashCode):
        return None
        pass

    def __checkConHashChange(self, lastConHashPrxList):
        tarsLogger.debug('AdapterProxyManager:checkConHashChange')
        lock = LockGuard(self.__newLock)
        if len(lastConHashPrxList) != len(self.__regAdapterProxyDict):
            return True
        regAdapterProxyList = sorted(
            list(self.__regAdapterProxyDict.items()), key=lambda item: item[0])
        regAdapterProxyListSize = len(regAdapterProxyList)
        for index in range(regAdapterProxyListSize):
            if cmp(lastConHashPrxList[index][0], regAdapterProxyList[index][0]) != 0:
                return True
        return False

    def __updateConHashProxyWeighted(self):
        tarsLogger.debug('AdapterProxyManager:updateConHashProxyWeighted')
        lock = LockGuard(self.__newLock)
        if len(self.__regAdapterProxyDict) == 0:
            raise TarsException("the adapter proxy is empty")
        self.__lastConHashPrxList = sorted(
            list(self.__regAdapterProxyDict.items()), key=lambda item: item[0])
        nodes = []
        for var in self.__lastConHashPrxList:
            nodes.append(var[0])
        if self.__consistentHashWeight is None:
            self.__consistentHashWeight = ConsistentHashNew(nodes)
        else:
            theOldActiveNodes = [
                var for var in nodes if var in self.__consistentHashWeight.nodes]

            theOldInactiveNodes = [
                var for var in self.__consistentHashWeight.nodes if var not in theOldActiveNodes]
            for var in theOldInactiveNodes:
                self.__consistentHashWeight.removeNode(var)

            theNewActiveNodes = [
                var for var in nodes if var not in theOldActiveNodes]
            for var in theNewActiveNodes:
                self.__consistentHashWeight.addNode(var)

            self.__consistentHashWeight.nodes = nodes
        pass

    def __getWeightedProxy(self):
        tarsLogger.debug('AdapterProxyManager:getWeightedProxy')
        lock = LockGuard(self.__newLock)
        if len(self.__adps) == 0:
            raise TarsException("the activate adapter proxy is empty")

        if self.__update is True:
            self.__lastWeightedProxyData.clear()
            weightedProxyData = {}
            minWeight = (list(self.__adps.items())[0][1][0]).getWeight()
            for item in list(self.__adps.items()):
                weight = (item[1][0].getWeight())
                weightedProxyData[item[0]] = (weight)
                if minWeight > weight:
                    minWeight = weight

            if minWeight <= 0:
                addWeight = -minWeight + 1
                for item in list(weightedProxyData.items()):
                    item[1] += addWeight

            self.__update = False
            self.__lastWeightedProxyData = weightedProxyData

        weightedProxyData = self.__lastWeightedProxyData
        while len(weightedProxyData) > 0:
            total = sum(weightedProxyData.values())
            rand = random.randint(1, total)
            temp = 0
            for item in list(weightedProxyData.items()):
                temp += item[1]
                if rand <= temp:
                    if self.__adps[item[0]][1].checkActive():
                        self.__adps[item[0]][2] += 1
                        return self.__adps[item[0]][1]
                    else:
                        weightedProxyData.pop(item[0])
                        break
        # 没有一个活跃的节点
        # 随机重连一个可用节点
        adpPrx = list(self.__adps.items())[random.randint(
            0, len(self.__adps))][1][1]
        adpPrx.checkActive()
        return None

    def selectAdapterProxy(self, reqmsg):
        '''
        @brief: 刷新本地缓存列表，如果服务下线了，要求删除本地缓存，通过一定算法返回AdapterProxy
        @param: reqmsg:请求响应报文
        @type reqmsg: ReqMessage
        @return:
        @rtype: EndPointInfo列表
        @todo: 优化负载均衡算法
        '''
        tarsLogger.debug('AdapterProxyManager:selectAdapterProxy')
        self.refreshEndpoints()
        if reqmsg.isHash:
            return self.__getHashProxy(reqmsg)
        else:
            if self.__weightType == EndpointWeightType.E_STATIC_WEIGHT:
                return self.__getWeightedProxy()
            else:
                return self.__getNextValidProxy()



# Tencent is pleased to support the open source community by making Tars available.
#
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
#

import time

from .EndpointF import EndpointF
from .__async import ServantProxyCallback
# from tars.core import tarscore;
# from tars.core import ServantProxy;
# from tars.core import ServantProxyCallback;
# from com.qq.register.EndpointF import *;
from .__init__ import tarscore
from .__servantproxy import ServantProxy


# proxy for client


class QueryFProxy(ServantProxy):
    def findObjectById(self, id, context=ServantProxy.mapcls_context()):
        oos = tarscore.TarsOutputStream()
        oos.write(tarscore.string, 1, id)

        rsp = self.tars_invoke(ServantProxy.TARSNORMAL,
                               "findObjectById", oos.getBuffer(), context, None)

        ios = tarscore.TarsInputStream(rsp.sBuffer)
        ret = ios.read(tarscore.vctclass(EndpointF), 0, True)

        return (ret)

    def async_findObjectById(self, callback, id, context=ServantProxy.mapcls_context()):
        oos = tarscore.TarsOutputStream()
        oos.write(tarscore.string, 1, id)

        self.tars_invoke_async(
            ServantProxy.TARSNORMAL, "findObjectById", oos.getBuffer(), context, None, callback)

    def findObjectById4Any(self, id, context=ServantProxy.mapcls_context()):
        oos = tarscore.TarsOutputStream()
        oos.write(tarscore.string, 1, id)

        rsp = self.tars_invoke(
            ServantProxy.TARSNORMAL, "findObjectById4Any", oos.getBuffer(), context, None)

        ios = tarscore.TarsInputStream(rsp.sBuffer)
        ret = ios.read(tarscore.int32, 0, True)
        activeEp = ios.read(tarscore.vctclass(EndpointF), 2, True)
        inactiveEp = ios.read(tarscore.vctclass(EndpointF), 3, True)

        return (ret, activeEp, inactiveEp)

    def async_findObjectById4Any(self, callback, id, context=ServantProxy.mapcls_context()):
        oos = tarscore.TarsOutputStream()
        oos.write(tarscore.string, 1, id)

        self.tars_invoke_async(
            ServantProxy.TARSNORMAL, "findObjectById4Any", oos.getBuffer(), context, None, callback)

    def findObjectById4All(self, id, context=ServantProxy.mapcls_context()):
        oos = tarscore.TarsOutputStream()
        oos.write(tarscore.string, 1, id)

        rsp = self.tars_invoke(
            ServantProxy.TARSNORMAL, "findObjectById4All", oos.getBuffer(), context, None)

        ios = tarscore.TarsInputStream(rsp.sBuffer)
        ret = ios.read(tarscore.int32, 0, True)
        activeEp = ios.read(tarscore.vctclass(EndpointF), 2, True)
        inactiveEp = ios.read(tarscore.vctclass(EndpointF), 3, True)

        return (ret, activeEp, inactiveEp)

    def async_findObjectById4All(self, callback, id, context=ServantProxy.mapcls_context()):
        oos = tarscore.TarsOutputStream()
        oos.write(tarscore.string, 1, id)

        self.tars_invoke_async(
            ServantProxy.TARSNORMAL, "findObjectById4All", oos.getBuffer(), context, None, callback)

    def findObjectByIdInSameGroup(self, id, context=ServantProxy.mapcls_context()):
        oos = tarscore.TarsOutputStream()
        oos.write(tarscore.string, 1, id)
        rsp = self.tars_invoke(
            ServantProxy.TARSNORMAL, "findObjectByIdInSameGroup", oos.getBuffer(), context, None)

        startDecodeTime = time.time()
        ios = tarscore.TarsInputStream(rsp.sBuffer)
        ret = ios.read(tarscore.int32, 0, True)
        activeEp = ios.read(tarscore.vctclass(EndpointF), 2, True)
        inactiveEp = ios.read(tarscore.vctclass(EndpointF), 3, True)
        endDecodeTime = time.time()
        return (ret, activeEp, inactiveEp, (endDecodeTime - startDecodeTime))

    def async_findObjectByIdInSameGroup(self, callback, id, context=ServantProxy.mapcls_context()):
        oos = tarscore.TarsOutputStream()
        oos.write(tarscore.string, 1, id)

        self.tars_invoke_async(
            ServantProxy.TARSNORMAL, "findObjectByIdInSameGroup", oos.getBuffer(), context, None, callback)

    def findObjectByIdInSameStation(self, id, sStation, context=ServantProxy.mapcls_context()):
        oos = tarscore.TarsOutputStream()
        oos.write(tarscore.string, 1, id)
        oos.write(tarscore.string, 2, sStation)

        rsp = self.tars_invoke(
            ServantProxy.TARSNORMAL, "findObjectByIdInSameStation", oos.getBuffer(), context, None)

        ios = tarscore.TarsInputStream(rsp.sBuffer)
        ret = ios.read(tarscore.int32, 0, True)
        activeEp = ios.read(tarscore.vctclass(EndpointF), 3, True)
        inactiveEp = ios.read(tarscore.vctclass(EndpointF), 4, True)

        return (ret, activeEp, inactiveEp)

    def async_findObjectByIdInSameStation(self, callback, id, sStation, context=ServantProxy.mapcls_context()):
        oos = tarscore.TarsOutputStream()
        oos.write(tarscore.string, 1, id)
        oos.write(tarscore.string, 2, sStation)

        self.tars_invoke_async(
            ServantProxy.TARSNORMAL, "findObjectByIdInSameStation", oos.getBuffer(), context, None, callback)

    def findObjectByIdInSameSet(self, id, setId, context=ServantProxy.mapcls_context()):
        oos = tarscore.TarsOutputStream()
        oos.write(tarscore.string, 1, id)
        oos.write(tarscore.string, 2, setId)

        rsp = self.tars_invoke(
            ServantProxy.TARSNORMAL, "findObjectByIdInSameSet", oos.getBuffer(), context, None)

        ios = tarscore.TarsInputStream(rsp.sBuffer)
        ret = ios.read(tarscore.int32, 0, True)
        activeEp = ios.read(tarscore.vctclass(EndpointF), 3, True)
        inactiveEp = ios.read(tarscore.vctclass(EndpointF), 4, True)

        return (ret, activeEp, inactiveEp)

    def async_findObjectByIdInSameSet(self, callback, id, setId, context=ServantProxy.mapcls_context()):
        oos = tarscore.TarsOutputStream()
        oos.write(tarscore.string, 1, id)
        oos.write(tarscore.string, 2, setId)

        self.tars_invoke_async(
            ServantProxy.TARSNORMAL, "findObjectByIdInSameSet", oos.getBuffer(), context, None, callback)


# ========================================================
# callback of async proxy for client
# ========================================================
class QueryFPrxCallback(ServantProxyCallback):
    def __init__(self):
        ServantProxyCallback.__init__(self)
        self.callback_map = {
            'findObjectById': self.__invoke_findObjectById,
            'findObjectById4Any': self.__invoke_findObjectById4Any,
            'findObjectById4All': self.__invoke_findObjectById4All,
            'findObjectByIdInSameGroup': self.__invoke_findObjectByIdInSameGroup,
            'findObjectByIdInSameStation': self.__invoke_findObjectByIdInSameStation,
            'findObjectByIdInSameSet': self.__invoke_findObjectByIdInSameSet
        }

    def callback_findObjectById(self, ret):
        raise NotImplementedError()

    def callback_findObjectById_exception(self, ret):
        raise NotImplementedError()

    def callback_findObjectById4Any(self, ret, activeEp, inactiveEp):
        raise NotImplementedError()

    def callback_findObjectById4Any_exception(self, ret):
        raise NotImplementedError()

    def callback_findObjectById4All(self, ret, activeEp, inactiveEp):
        raise NotImplementedError()

    def callback_findObjectById4All_exception(self, ret):
        raise NotImplementedError()

    def callback_findObjectByIdInSameGroup(self, ret, activeEp, inactiveEp):
        raise NotImplementedError()

    def callback_findObjectByIdInSameGroup_exception(self, ret):
        raise NotImplementedError()

    def callback_findObjectByIdInSameStation(self, ret, activeEp, inactiveEp):
        raise NotImplementedError()

    def callback_findObjectByIdInSameStation_exception(self, ret):
        raise NotImplementedError()

    def callback_findObjectByIdInSameSet(self, ret, activeEp, inactiveEp):
        raise NotImplementedError()

    def callback_findObjectByIdInSameSet_exception(self, ret):
        raise NotImplementedError()

    def __invoke_findObjectById(self, reqmsg):
        rsp = reqmsg.response
        if rsp.iRet != ServantProxy.TARSSERVERSUCCESS:
            self.callback_findObjectById_exception(rsp.iRet)
            return rsp.iRet
        ios = tarscore.TarsInputStream(rsp.sBuffer)
        ret = ios.read(tarscore.vctclass(EndpointF), 0, True)
        self.callback_findObjectById(ret)

    def __invoke_findObjectById4Any(self, reqmsg):
        rsp = reqmsg.response
        if rsp.iRet != ServantProxy.TARSSERVERSUCCESS:
            self.callback_findObjectById4Any_exception(rsp.iRet)
            return rsp.iRet
        ios = tarscore.TarsInputStream(rsp.sBuffer)
        ret = ios.read(tarscore.int32, 0, True)
        activeEp = ios.read(tarscore.vctclass(EndpointF), 2, True)
        inactiveEp = ios.read(tarscore.vctclass(EndpointF), 3, True)
        self.callback_findObjectById4Any(ret, activeEp, inactiveEp)

    def __invoke_findObjectById4All(self, reqmsg):
        rsp = reqmsg.response
        if rsp.iRet != ServantProxy.TARSSERVERSUCCESS:
            self.callback_findObjectById4All_exception(rsp.iRet)
            return rsp.iRet
        ios = tarscore.TarsInputStream(rsp.sBuffer)
        ret = ios.read(tarscore.int32, 0, True)
        activeEp = ios.read(tarscore.vctclass(EndpointF), 2, True)
        inactiveEp = ios.read(tarscore.vctclass(EndpointF), 3, True)
        self.callback_findObjectById4All(ret, activeEp, inactiveEp)

    def __invoke_findObjectByIdInSameGroup(self, reqmsg):
        rsp = reqmsg.response
        if rsp.iRet != ServantProxy.TARSSERVERSUCCESS:
            self.callback_findObjectByIdInSameGroup_exception(rsp.iRet)
            return rsp.iRet
        ios = tarscore.TarsInputStream(rsp.sBuffer)
        ret = ios.read(tarscore.int32, 0, True)
        activeEp = ios.read(tarscore.vctclass(EndpointF), 2, True)
        inactiveEp = ios.read(tarscore.vctclass(EndpointF), 3, True)
        self.callback_findObjectByIdInSameGroup(ret, activeEp, inactiveEp)

    def __invoke_findObjectByIdInSameStation(self, reqmsg):
        rsp = reqmsg.response
        if rsp.iRet != ServantProxy.TARSSERVERSUCCESS:
            self.callback_findObjectByIdInSameStation_exception(rsp.iRet)
            return rsp.iRet
        ios = tarscore.TarsInputStream(rsp.sBuffer)
        ret = ios.read(tarscore.int32, 0, True)
        activeEp = ios.read(tarscore.vctclass(EndpointF), 3, True)
        inactiveEp = ios.read(tarscore.vctclass(EndpointF), 4, True)
        self.callback_findObjectByIdInSameStation(ret, activeEp, inactiveEp)

    def __invoke_findObjectByIdInSameSet(self, reqmsg):
        rsp = reqmsg.response
        if rsp.iRet != ServantProxy.TARSSERVERSUCCESS:
            self.callback_findObjectByIdInSameSet_exception(rsp.iRet)
            return rsp.iRet
        ios = tarscore.TarsInputStream(rsp.sBuffer)
        ret = ios.read(tarscore.int32, 0, True)
        activeEp = ios.read(tarscore.vctclass(EndpointF), 3, True)
        inactiveEp = ios.read(tarscore.vctclass(EndpointF), 4, True)
        self.callback_findObjectByIdInSameSet(ret, activeEp, inactiveEp)

    def onDispatch(self, reqmsg):
        self.callback_map[reqmsg.request.sFuncName](reqmsg)



import re, traceback, datetime, base64
import asyncio
from biliup.plugins import random_user_agent

# The core codes for YouTube support are basically from taizan-hokuto/pytchat

headers = {
    'user-agent': random_user_agent(),
}


class Youtube:
    q = None
    url = ""
    vid = ""
    ctn = ""
    client = None
    stop = False

    @classmethod
    async def run(cls, url, q, client, **kargs):
        from .paramgen import liveparam

        cls.q = q
        cls.url = url
        cls.client = client
        cls.stop = False
        cls.key = "eW91dHViZWkvdjEvbGl2ZV9jaGF0L2dldF9saXZlX2NoYXQ/a2V5PUFJemFTeUFPX0ZKMlNscVU4UTRTVEVITEdDaWx3X1k5XzExcWNXOA=="
        await cls.get_url()
        while cls.stop == False:
            try:
                await cls.get_room_info()
                cls.ctn = liveparam.getparam(cls.vid, cls.cid, 1)
                await cls.get_chat()
            except:
                traceback.print_exc()
                await asyncio.sleep(1)

    @classmethod
    async def stop(cls):
        cls.stop = True

    @classmethod
    async def get_url(cls):
        a = re.search(r"youtube.com/channel/([^/?]+)", cls.url)
        try:
            cid = a.group(1)
            cls.cid = cid
            cls.url = f"https://www.youtube.com/channel/{cid}/videos"
        except:
            a = re.search(r"youtube.com/watch\?v=([^/?]+)", cls.url)
            async with cls.client.request(
                "get", f"https://www.youtube.com/embed/{a.group(1)}"
            ) as resp:
                b = re.search(r'\\"channelId\\":\\"(.{24})\\"', await resp.text())
                cls.cid = b.group(1)
                cls.url = f"https://www.youtube.com/channel/{cls.cid}/videos"

    @classmethod
    async def get_room_info(cls):
        async with cls.client.request("get", cls.url) as resp:
            t = re.search(
                r'"gridVideoRenderer"((.(?!"gridVideoRenderer"))(?!"style":"UPCOMING"))+"label":"(LIVE|LIVE NOW|PREMIERING NOW)"([\s\S](?!"style":"UPCOMING"))+?("gridVideoRenderer"|</script>)',
                await resp.text(),
            ).group(0)
            cls.vid = re.search(r'"gridVideoRenderer".+?"videoId":"(.+?)"', t).group(1)
            # print(cls.vid)

    @classmethod
    async def get_chat_single(cls):
        msgs = []
        data = {
            "context": {
                "client": {
                    "visitorData": "",
                    "userAgent": headers["user-agent"],
                    "clientName": "WEB",
                    "clientVersion": "".join(
                        (
                            "2.",
                            (datetime.datetime.today() - datetime.timedelta(days=1)).strftime(
                                "%Y%m%d"
                            ),
                            ".01.00",
                        )
                    ),
                },
            },
            "continuation": cls.ctn,
        }
        u = f'https://www.youtube.com/{base64.b64decode(cls.key).decode("utf-8")}'
        async with cls.client.request("post", u, headers=headers, json=data) as resp:
            # print(await resp.text())
            j = await resp.json()
            j = j["continuationContents"]
            cont = j["liveChatContinuation"]["continuations"][0]
            if cont is None:
                raise Exception("No Continuation")
            metadata = (
                cont.get("invalidationContinuationData")
                or cont.get("timedContinuationData")
                or cont.get("reloadContinuationData")
                or cont.get("liveChatReplayContinuationData")
            )
            cls.ctn = metadata["continuation"]
            # print(j['liveChatContinuation'].get('actions'))
            for action in j["liveChatContinuation"].get("actions", []):
                try:
                    renderer = action["addChatItemAction"]["item"]["liveChatTextMessageRenderer"]
                    msg = {}
                    msg["name"] = renderer["authorName"]["simpleText"]
                    message = ""
                    runs = renderer["message"].get("runs")
                    for r in runs:
                        if r.get("emoji"):
                            message += r["emoji"].get("shortcuts", [""])[0]
                        else:
                            message += r.get("text", "")
                    msg["content"] = message
                    msg["msg_type"] = "danmaku"
                    msg["color"] = "16777215"
                    msgs.append(msg)
                except:
                    pass

        return msgs

    @classmethod
    async def get_chat(cls):
        while cls.stop == False:
            ms = await cls.get_chat_single()
            if len(ms) != 0:
                interval = 1 / len(ms)
            else:
                await asyncio.sleep(1)
            for m in ms:
                await cls.q.put(m)
                await asyncio.sleep(interval)



from biliup.common.tars import tarscore

class HuyaStreamInfo(tarscore.struct):

    __tars_class__ = "Huya.StreamInfo"

    def __init__(self):
        self.sCdnType: tarscore.string = ""
        self.iIsMaster: tarscore.int32 = 0
        self.lChannelId: tarscore.int64 = 0
        self.lSubChannelId: tarscore.int64 = 0
        self.lPresenterUid: tarscore.int64 = 0
        self.sStreamName: tarscore.string = ""
        self.sFlvUrl: tarscore.string = ""
        self.sFlvUrlSuffix: tarscore.string = ""
        self.sFlvAntiCode: tarscore.string = ""
        self.sHlsUrl: tarscore.string = ""
        self.sHlsUrlSuffix: tarscore.string = ""
        self.sHlsAntiCode: tarscore.string = ""
        self.iLineIndex: tarscore.int32 = 0
        self.iIsMultiStream: tarscore.int32 = 0
        self.iPCPriorityRate: tarscore.int32 = 0
        self.iWebPriorityRate: tarscore.int32 = 0
        self.iMobilePriorityRate: tarscore.int32 = 0

    @staticmethod
    def writeTo(oos: tarscore.TarsOutputStream, value):
        oos.write(tarscore.string, 0, value.sCdnType)
        oos.write(tarscore.int32, 1, value.iIsMaster)
        oos.write(tarscore.int64, 2, value.lChannelId)
        oos.write(tarscore.int64, 3, value.lSubChannelId)
        oos.write(tarscore.int64, 4, value.lPresenterUid)
        oos.write(tarscore.string, 5, value.sStreamName)
        oos.write(tarscore.string, 6, value.sFlvUrl)
        oos.write(tarscore.string, 7, value.sFlvUrlSuffix)
        oos.write(tarscore.string, 8, value.sFlvAntiCode)
        oos.write(tarscore.string, 9, value.sHlsUrl)
        oos.write(tarscore.string, 10, value.sHlsUrlSuffix)
        oos.write(tarscore.string, 11, value.sHlsAntiCode)
        oos.write(tarscore.int32, 12, value.iLineIndex)
        oos.write(tarscore.int32, 13, value.iIsMultiStream)
        oos.write(tarscore.int32, 14, value.iPCPriorityRate)
        oos.write(tarscore.int32, 15, value.iWebPriorityRate)
        oos.write(tarscore.int32, 16, value.iMobilePriorityRate)

    @staticmethod
    def readFrom(ios: tarscore.TarsInputStream):
        value = HuyaStreamInfo()
        value.sCdnType = ios.read(tarscore.string, 0, False)
        value.iIsMaster = ios.read(tarscore.int32, 1, False)
        value.lChannelId = ios.read(tarscore.int64, 2, False)
        value.lSubChannelId = ios.read(tarscore.int64, 3, False)
        value.lPresenterUid = ios.read(tarscore.int64, 4, False)
        value.sStreamName = ios.read(tarscore.string, 5, False)
        value.sFlvUrl = ios.read(tarscore.string, 6, False)
        value.sFlvUrlSuffix = ios.read(tarscore.string, 7, False)
        value.sFlvAntiCode = ios.read(tarscore.string, 8, False)
        value.sHlsUrl = ios.read(tarscore.string, 9, False)
        value.sHlsUrlSuffix = ios.read(tarscore.string, 10, False)
        value.sHlsAntiCode = ios.read(tarscore.string, 11, False)
        value.iLineIndex = ios.read(tarscore.int32, 12, False)
        value.iIsMultiStream = ios.read(tarscore.int32, 13, False)
        value.iPCPriorityRate = ios.read(tarscore.int32, 14, False)
        value.iWebPriorityRate = ios.read(tarscore.int32, 15, False)
        value.iMobilePriorityRate = ios.read(tarscore.int32, 16, False)
        return value






# 部分弹幕功能代码来自项目：https://github.com/IsoaSFlus/danmaku，感谢大佬
# 快手弹幕代码来源及思路：https://github.com/py-wuhao/ks_barrage，感谢大佬
# 部分斗鱼录播修复代码与思路来源于：https://github.com/SmallPeaches/DanmakuRender，感谢大佬
# 仅抓取用户弹幕，不包括入场提醒、礼物赠送等。

import asyncio
import logging
import os
import re
import ssl
import threading
import time
from abc import ABC, abstractmethod
from typing import Optional

import aiohttp
import lxml.etree as etree

from biliup.Danmaku.bilibili import Bilibili
from biliup.Danmaku.douyin import Douyin
from biliup.Danmaku.douyu import Douyu
from biliup.Danmaku.huya import Huya
from biliup.Danmaku.twitcasting import Twitcasting
from biliup.Danmaku.twitch import Twitch

logger = logging.getLogger('biliup')


class IDanmakuClient(ABC):
    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def save(self, file_name: Optional[str] = None):
        pass


class DanmakuClient(IDanmakuClient):
    class WebsocketErrorException(Exception):
        pass

    def __init__(self, url, file_name, content=None):
        # TODO 录制任务产生的上下文信息 传递太麻烦了 需要改
        self.__content = content if content is not None else {}
        self.__file_name = file_name
        self.__fmt_file_name = None
        self.__url = ''
        self.__site = None
        self.__hs = None
        self.__ws = None
        self.__dm_queue: Optional[asyncio.Queue] = None
        self.__record_task: Optional[asyncio.Task] = None
        self.__print_task: Optional[asyncio.Task] = None

        if 'http://' == url[:7] or 'https://' == url[:8]:
            self.__url = url
        else:
            self.__url = 'http://' + url
        for u, s in {'douyu.com': Douyu,
                     'huya.com': Huya,
                     'live.bilibili.com': Bilibili,
                     'twitch.tv': Twitch,
                     'douyin.com': Douyin,
                     'twitcasting.tv': Twitcasting
                     }.items():
            if re.match(r'^(?:http[s]?://)?.*?%s/(.+?)$' % u, url):
                self.__site = s
                self.__u = u
                break

        if self.__site is None:
            # 抛出异常由外部处理 exit()会导致进程退出
            raise Exception(f"{DanmakuClient.__name__}:{self.__url}: 不支持录制弹幕")

    async def __init_ws(self):
        try:
            ws_url, reg_datas = await self.__site.get_ws_info(self.__url, self.__content)
            ctx = ssl.create_default_context()
            ctx.set_ciphers('DEFAULT')
            self.__ws = await self.__hs.ws_connect(ws_url, ssl_context=ctx, headers=getattr(self.__site, 'headers', {}))
            for reg_data in reg_datas:
                if type(reg_data) == str:
                    await self.__ws.send_str(reg_data)
                else:
                    await self.__ws.send_bytes(reg_data)
        except asyncio.CancelledError:
            raise
        except:
            raise self.WebsocketErrorException()

    async def __heartbeats(self):
        if self.__site.heartbeat is not None:
            while self.__site.heartbeat:
                # 每隔这么长时间发送一次心跳包
                await asyncio.sleep(self.__site.heartbeatInterval)
                # 发送心跳包
                if type(self.__site.heartbeat) == str:
                    await self.__ws.send_str(self.__site.heartbeat)
                else:
                    await self.__ws.send_bytes(self.__site.heartbeat)

    async def __fetch_danmaku(self):
        while True:
            # 使用 async for msg in self.__ws
            # 会导致在连接断开时 需要很长时间15min或者更多才能检测到
            msg = await self.__ws.receive()

            if msg.type in [aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.ERROR]:
                # 连接关闭的异常 会到外层统一处理
                raise self.WebsocketErrorException()

            try:
                result = self.__site.decode_msg(msg.data)

                if isinstance(result, tuple):
                    ms, ack = result
                    if ack is not None:
                        # 发送ack包
                        if type(ack) == str:
                            await self.__ws.send_str(ack)
                        else:
                            await self.__ws.send_bytes(ack)
                else:
                    ms = result

                for m in ms:
                    await self.__dm_queue.put(m)
            except asyncio.CancelledError:
                raise
            except:
                logger.exception(f"{DanmakuClient.__name__}:{self.__url}: 弹幕接收异常")
                continue
                # await asyncio.sleep(10) 无需等待 直接获取下一条websocket消息
                # 这里出现异常只会是 decode_msg 的问题

    async def __print_danmaku(self):
        def write_file(filename):
            try:
                if filename and msg_i > 0:
                    tree.write(filename, encoding="UTF-8", xml_declaration=True, pretty_print=True)
            except:
                logger.warning(f"{DanmakuClient.__name__}:{self.__url}: 弹幕写入异常", exc_info=True)

        while True:
            root = etree.Element("i")
            etree.indent(root, "\t")
            tree = etree.ElementTree(root, parser=etree.XMLParser(recover=True))
            start_time = time.time()
            fmt_file_name = time.strftime(self.__file_name.encode("unicode-escape").decode()).encode().decode(
                "unicode-escape") + '.xml'
            msg_i = 0

            # 设置弹幕自动保存时间
            last_save_time = int(start_time)
            if self.__content.get("raw", False):
                save_interval = 300
            else:
                save_interval = 10

            try:
                while True:
                    try:
                        # 无弹幕时更快分段结束
                        m = await asyncio.wait_for(self.__dm_queue.get(), timeout=1)
                    except asyncio.TimeoutError:
                        continue

                    msg_type = m.get("msg_type")
                    msg_time = time.time()
                    logger.debug(f"{DanmakuClient.__name__}:{self.__url}: 弹幕queue-{msg_type}")
                    # print(m)
                    # 非弹幕
                    if msg_type == "save":
                        if 'file_name' in m and fmt_file_name != m['file_name']:
                            try:
                                if os.path.exists(m['file_name']):
                                    os.remove(m['file_name'])
                                if os.path.exists(fmt_file_name):
                                    os.rename(fmt_file_name, m.get('file_name'))
                                    logger.info(
                                        f"{DanmakuClient.__name__}:{self.__url}: 更名 {fmt_file_name} 为 {m['file_name']}")
                            except:
                                logger.exception(
                                    f"{DanmakuClient.__name__}:{self.__url}: 更名 {fmt_file_name} 为 {m['file_name']}失败")
                            fmt_file_name = m['file_name']

                        if callable(m.get('callback')):
                            m['callback']()
                        break
                    elif msg_type == "stop":
                        try:
                            os.remove(fmt_file_name)
                        except:
                            pass
                        fmt_file_name = None
                        self.__record_task.cancel()
                        return
                    else: # 正常弹幕
                        # 完整弹幕记录
                        if self.__content.get("raw", False):
                            try:
                                o = etree.SubElement(root, 'o')
                                o.set('timestamp', str(int(msg_time)))
                                o.text = m.get("raw_data")
                            except:
                                logger.warning(f"{DanmakuClient.__name__}:{self.__url}:弹幕处理异常", exc_info=True)
                                continue

                        # 正常弹幕记录
                        if msg_type == 'danmaku': # 文字弹幕
                            try:
                                if m.get('color'):
                                    color = m["color"]
                                else:
                                    color = '16777215'
                                msg_time_since_start = format(msg_time - start_time, '.3f')
                                # 记录弹幕额外信息
                                timestamp = str(int(msg_time))
                                uid = str(m.get("uid",0))
                                d = etree.SubElement(root, 'd')
                                '''
                                弹幕部分的参数兼容bilibili主站弹幕 XML 文件，按顺序含义分别为：
                                1.弹幕出现时间 (秒)
                                2.弹幕类型
                                3.字号
                                4.颜色
                                5.发送时间戳
                                6.固定为 0 (主站弹幕 XML 为弹幕池 ID)
                                7.发送者 UID (主站弹幕 XML 为发送用户 ID 的 CRC32)
                                8.固定为 0 (主站弹幕 XML 为弹幕的数据库 ID)
                                '''
                                d.set('p', f"{msg_time_since_start},1,25,{color},{timestamp},0,{uid},0")
                                if self.__content.get("detail", False):
                                    d.set('timestamp', timestamp)
                                    d.set('uid',uid)
                                    #兼容DanmakuFactory用户名识别
                                    d.set('user', m.get("name",""))
                                d.text = m["content"]
                            except:
                                logger.warning(f"{DanmakuClient.__name__}:{self.__url}:弹幕处理异常", exc_info=True)
                                # 异常后略过本次弹幕
                                continue
                        # 礼物信息记录，支持上舰、SC、礼物，目前仅在B站开启
                        elif self.__u == 'live.bilibili.com' and msg_type in ['gift', 'super_chat' , 'guard_buy']:
                            if not self.__content.get("detail", False):
                                continue
                            try:
                                s = etree.SubElement(root, 's')
                                s.set('timestamp', str(int(msg_time)))
                                s.set('uid', str(m.get("uid")))
                                s.set('username', m.get("name",""))
                                s.set('price', str(m.get("price")))
                                s.set('type', msg_type)
                                # 礼物名称
                                s.set('num', str(m.get('num')))
                                s.set('giftname', m.get('gift_name'))
                                s.text = m["content"]
                            except:
                                logger.warning(f"{DanmakuClient.__name__}:{self.__url}:弹幕处理异常", exc_info=True)
                                # 异常后略过本次弹幕
                                continue
                        else:
                            # 如果未开启了原始弹幕，则跳过本次循环
                            if not self.__content.get("raw", False):
                                continue
                        msg_i += 1

                    # 每隔指定时间接入一次弹幕
                    time_since_last_save = int(msg_time) - last_save_time
                    if msg_i > 0 and time_since_last_save >= save_interval:
                        logger.debug(f'{DanmakuClient.__name__}:{self.__url}: 写入弹幕')
                        write_file(fmt_file_name)
                        last_save_time = int(msg_time)
                        msg_i = 0
                    # else: # 弹幕未写入原因
                    #     print(f"time_since_last_save: {time_since_last_save}, msg_i: {msg_i}")
            finally:
                # 发生异常(被取消)时写入 避免丢失未写入
                logger.debug(f'{DanmakuClient.__name__}:{self.__url}:弹幕保底写入')
                write_file(fmt_file_name)

    def start(self):
        init_event = threading.Event()

        async def __init():
            logger.info(f'开始弹幕录制: {self.__url}')
            self.__record_task = asyncio.create_task(self.__run())
            init_event.set()
            try:
                await self.__record_task
            except asyncio.CancelledError:
                pass
            self.__record_task = None
            logger.info(f'结束弹幕录制: {self.__url}')

        threading.Thread(target=asyncio.run, args=(__init(),)).start()
        # 等待初始化完成避免未初始化完成的时候就停止任务
        init_event.wait()

    def save(self, file_name: Optional[str] = None):
        if self.__record_task:
            logger.debug(f"{DanmakuClient.__name__}:{self.__url}: 弹幕save")
            init_event = threading.Event()
            self.__dm_queue.put_nowait({
                "msg_type": "save",
                "file_name": file_name,
                "callback": lambda: init_event.set()
            })
            init_event.wait()

    def stop(self):
        if self.__record_task:
            logger.debug(f"{DanmakuClient.__name__}:{self.__url}: 弹幕stop")
            self.__dm_queue.put_nowait({
                "msg_type": "stop",
            })

    async def __run(self):
        try:
            self.__dm_queue = asyncio.Queue()
            self.__hs = aiohttp.ClientSession()
            self.__print_task = asyncio.create_task(self.__print_danmaku())
            while True:
                danmaku_tasks = []
                try:
                    await self.__init_ws()
                    danmaku_tasks = [asyncio.create_task(self.__heartbeats()),
                                     asyncio.create_task(self.__fetch_danmaku())]
                    await asyncio.gather(*danmaku_tasks)
                except asyncio.CancelledError:
                    raise
                except self.WebsocketErrorException:
                    # 连接断开等30秒重连
                    # 在关闭之前一直重试
                    if self.__u == 'live.bilibili.com' and self.__content['uid'] != 0:
                        self.__content['uid'] = 0
                        logger.warning(f"{DanmakuClient.__name__}:{self.__url}: 弹幕连接异常,降级至非完整弹幕")
                        continue
                    logger.warning(f"{DanmakuClient.__name__}:{self.__url}: 弹幕连接异常,将在 30 秒后重试", exc_info=True)
                except:
                    # 记录异常不到外部处理
                    logger.exception(f"{DanmakuClient.__name__}:{self.__url}: 弹幕异常,将在 30 秒后重试")
                finally:
                    if danmaku_tasks:
                        for danmaku_task in danmaku_tasks:
                            danmaku_task.cancel()
                        await asyncio.wait(danmaku_tasks)
                    if self.__ws is not None and not self.__ws.closed:
                        await self.__ws.close()
                await asyncio.sleep(30)
        finally:
            if self.__print_task:
                self.__print_task.cancel()
                await asyncio.wait([self.__print_task])
            if self.__hs:
                await self.__hs.close()

# 虎牙直播：https://www.huya.com/lpl
# 斗鱼直播：https://www.douyu.com/9999



from biliup.common.tars import tarscore
from .StreamInfo import HuyaStreamInfo
from .MultiStreamInfo import HuyaMultiStreamInfo


class HuyaBeginLiveNotice(tarscore.struct):

    __tars_class__ = "Huya.BeginLiveNotice"

    def __init__(self):
        self.lPresenterUid: tarscore.int64 = 0
        self.iGameId: tarscore.int32 = 0
        self.sGameName: tarscore.string = ""
        self.iRandomRange: tarscore.int32 = 0
        self.iStreamType: tarscore.int32 = 0
        self.vStreamInfo: tarscore.vector = tarscore.vector(HuyaStreamInfo) # FIXME
        self.vCdnList: tarscore.vector = tarscore.vector(tarscore.string) # FIXME
        self.lLiveId: tarscore.int64 = 0
        self.iPCDefaultBitRate: tarscore.int32 = 0
        self.iWebDefaultBitRate: tarscore.int32 = 0
        self.iMobileDefaultBitRate: tarscore.int32 = 0
        self.lMultiStreamFlag: tarscore.int64 = 0
        self.sNick: tarscore.string = ""
        self.lYYId: tarscore.int64 = 0
        self.lAttendeeCount: tarscore.int64 = 0
        self.iCodecType: tarscore.int32 = 0
        self.iScreenType: tarscore.int32 = 0
        self.vMultiStreamInfo: tarscore.vector = tarscore.vector(HuyaMultiStreamInfo) # FIXME
        self.sLiveDesc: tarscore.string = ""
        self.lLiveCompatibleFlag: tarscore.int64 = 0
        self.sAvatarUrl: tarscore.string = ""
        self.iSourceType: tarscore.int32 = 0
        self.sSubchannelName: tarscore.string = ""
        self.sVideoCaptureUrl: tarscore.string = ""
        self.iStartTime: tarscore.int32 = 0
        self.lChannelId: tarscore.int64 = 0
        self.lSubChannelId: tarscore.int64 = 0
        self.sLocation: tarscore.string = ""

    @staticmethod
    def writeTo(oos: tarscore.TarsOutputStream, value):
        oos.write(tarscore.int64, 0, value.lPresenterUid)
        oos.write(tarscore.int32, 1, value.iGameId)
        oos.write(tarscore.string, 2, value.sGameName)
        oos.write(tarscore.int32, 3, value.iRandomRange)
        oos.write(tarscore.int32, 4, value.iStreamType)
        oos.write(tarscore.vector, 5, value.vStreamInfo)
        oos.write(tarscore.vector, 6, value.vCdnList)
        oos.write(tarscore.int64, 7, value.lLiveId)
        oos.write(tarscore.int32, 8, value.iPCDefaultBitRate)
        oos.write(tarscore.int32, 9, value.iWebDefaultBitRate)
        oos.write(tarscore.int32, 10, value.iMobileDefaultBitRate)
        oos.write(tarscore.int64, 11, value.lMultiStreamFlag)
        oos.write(tarscore.string, 12, value.sNick)
        oos.write(tarscore.int64, 13, value.lYYId)
        oos.write(tarscore.int64, 14, value.lAttendeeCount)
        oos.write(tarscore.int32, 15, value.iCodecType)
        oos.write(tarscore.int32, 16, value.iScreenType)
        oos.write(tarscore.vector, 17, value.vMultiStreamInfo)
        oos.write(tarscore.string, 18, value.sLiveDesc)
        oos.write(tarscore.int64, 19, value.lLiveCompatibleFlag)
        oos.write(tarscore.string, 20, value.sAvatarUrl)
        oos.write(tarscore.int32, 21, value.iSourceType)
        oos.write(tarscore.string, 22, value.sSubchannelName)
        oos.write(tarscore.string, 23, value.sVideoCaptureUrl)
        oos.write(tarscore.int32, 24, value.iStartTime)
        oos.write(tarscore.int64, 25, value.lChannelId)
        oos.write(tarscore.int64, 26, value.lSubChannelId)
        oos.write(tarscore.string, 27, value.sLocation)

    @staticmethod
    def readFrom(ios: tarscore.TarsInputStream):
        value = HuyaBeginLiveNotice()
        value.lPresenterUid = ios.read(tarscore.int64, 0, False)
        value.iGameId = ios.read(tarscore.int32, 1, False)
        value.sGameName = ios.read(tarscore.string, 2, False)
        value.iRandomRange = ios.read(tarscore.int32, 3, False)
        value.iStreamType = ios.read(tarscore.int32, 4, False)
        value.vStreamInfo = ios.read(tarscore.vector, 5, False)
        value.vCdnList = ios.read(tarscore.vector, 6, False)
        value.lLiveId = ios.read(tarscore.int64, 7, False)
        value.iPCDefaultBitRate = ios.read(tarscore.int32, 8, False)
        value.iWebDefaultBitRate = ios.read(tarscore.int32, 9, False)
        value.iMobileDefaultBitRate = ios.read(tarscore.int32, 10, False)
        value.lMultiStreamFlag = ios.read(tarscore.int64, 11, False)
        value.sNick = ios.read(tarscore.string, 12, False)
        value.lYYId = ios.read(tarscore.int64, 13, False)
        value.lAttendeeCount = ios.read(tarscore.int64, 14, False)
        value.iCodecType = ios.read(tarscore.int32, 15, False)
        value.iScreenType = ios.read(tarscore.int32, 16, False)
        value.vMultiStreamInfo = ios.read(tarscore.vector, 17, False)
        value.sLiveDesc = ios.read(tarscore.string, 18, False)
        value.lLiveCompatibleFlag = ios.read(tarscore.int64, 19, False)
        value.sAvatarUrl = ios.read(tarscore.string, 20, False)
        value.iSourceType = ios.read(tarscore.int32, 21, False)
        value.sSubchannelName = ios.read(tarscore.string, 22, False)
        value.sVideoCaptureUrl = ios.read(tarscore.string, 23, False)
        value.iStartTime = ios.read(tarscore.int32, 24, False)
        value.lChannelId = ios.read(tarscore.int64, 25, False)
        value.lSubChannelId = ios.read(tarscore.int64, 26, False)
        value.sLocation = ios.read(tarscore.string, 27, False)
        return value

    def as_dict(self):
        return {
            "lPresenterUid": self.lPresenterUid,
            "iGameId": self.iGameId,
            "sGameName": self.sGameName,
            "iRandomRange": self.iRandomRange,
            "iStreamType": self.iStreamType,
            "vStreamInfo": self.vStreamInfo,
            "vCdnList": self.vCdnList,
            "lLiveId": self.lLiveId,
            "iPCDefaultBitRate": self.iPCDefaultBitRate,
            "iWebDefaultBitRate": self.iWebDefaultBitRate,
            "iMobileDefaultBitRate": self.iMobileDefaultBitRate,
            "lMultiStreamFlag": self.lMultiStreamFlag,
            "sNick": self.sNick,
            "lYYId": self.lYYId,
            "lAttendeeCount": self.lAttendeeCount,
            "iCodecType": self.iCodecType,
            "iScreenType": self.iScreenType,
            "vMultiStreamInfo": self.vMultiStreamInfo,
            "sLiveDesc": self.sLiveDesc,
            "lLiveCompatibleFlag": self.lLiveCompatibleFlag,
            "sAvatarUrl": self.sAvatarUrl,
            "iSourceType": self.iSourceType,
            "sSubchannelName": self.sSubchannelName,
            "sVideoCaptureUrl": self.sVideoCaptureUrl,
            "iStartTime": self.iStartTime,
            "lChannelId": self.lChannelId,
            "lSubChannelId": self.lSubChannelId,
            "sLocation": self.sLocation
        }